"""
Source fitting routines.
"""

import math

import numpy
import scipy.optimize
from .gaussian import gaussian, jac_gaussian
from .stats import indep_pixels
from sourcefinder.deconv import deconv
from numba import guvectorize, float64, float32, int32

FIT_PARAMS = ('peak', 'xbar', 'ybar', 'semimajor', 'semiminor', 'theta')


def moments(data, fudge_max_pix_factor, beamsize, threshold=0):
    """Calculate source positional values using moments

    Use the first moment of the distribution is the barycenter of an
    ellipse. The second moments are used to estimate the rotation angle
    and the length of the axes.

    Args:

        data (numpy.ndarray): Actual 2D image data

        fudge_max_pix_factor(float): Correct for the underestimation of the peak
                                     by taking the maximum pixel value.

        beamsize(float): The FWHM size of the clean beam

        threshold(float): source parameters like the semimajor and semiminor axes
                          derived from moments can be underestimated if one does not take
                          account of the threshold that was used to segment the source islands.

    Returns:
        dict: peak, total, x barycenter, y barycenter, semimajor
            axis, semiminor axis, theta

    Raises:
        exceptions.ValueError: in case of NaN in input.

    """

    # Are we fitting a -ve or +ve Gaussian?
    if data.mean() >= 0:
        # The peak is always underestimated when you take the highest pixel.
        peak = data.max() * fudge_max_pix_factor
    else:
        peak = data.min()
    ratio = threshold / peak
    total = data.sum()
    x, y = numpy.indices(data.shape)
    xbar = float((x * data).sum() / total)
    ybar = float((y * data).sum() / total)
    xxbar = (x * x * data).sum() / total - xbar ** 2
    yybar = (y * y * data).sum() / total - ybar ** 2
    xybar = (x * y * data).sum() / total - xbar * ybar

    working1 = (xxbar + yybar) / 2.0
    working2 = math.sqrt(((xxbar - yybar) / 2) ** 2 + xybar ** 2)

    # Some problems arise with the sqrt of (working1-working2) when they are
    # equal, this happens with islands that have a thickness of only one pixel
    # in at least one dimension.  Due to rounding errors this difference
    # becomes negative--->math domain error in sqrt.
    if len(data.nonzero()[0]) == 1:
        # This is the case when the island (or more likely subisland) has
        # a size of only one pixel.
        semiminor = numpy.sqrt(beamsize / numpy.pi)
        semimajor = numpy.sqrt(beamsize / numpy.pi)
    else:
        semimajor_tmp = (working1 + working2) * 2.0 * math.log(2.0)
        semiminor_tmp = (working1 - working2) * 2.0 * math.log(2.0)
        # ratio will be 0 for data that hasn't been selected according to a
        # threshold.
        if ratio != 0:
            # The corrections below for the semi-major and semi-minor axes are
            # to compensate for the underestimate of these quantities
            # due to the cutoff at the threshold.
            semimajor_tmp /= (1.0 + math.log(ratio) * ratio / (1.0 - ratio))
            semiminor_tmp /= (1.0 + math.log(ratio) * ratio / (1.0 - ratio))
        semimajor = math.sqrt(semimajor_tmp)
        semiminor = math.sqrt(semiminor_tmp)
        if semiminor == 0:
            # A semi-minor axis exactly zero gives all kinds of problems.
            # For instance wrt conversion to celestial coordinates.
            # This is a quick fix.
            semiminor = beamsize / (numpy.pi * semimajor)

    if (numpy.isnan(xbar) or numpy.isnan(ybar) or
            numpy.isnan(semimajor) or numpy.isnan(semiminor)):
        raise ValueError("Unable to estimate Gauss shape")

    # Theta is not affected by the cut-off at the threshold (see Spreeuw 2010,
    # page 45).
    if abs(semimajor - semiminor) < 0.01:
        # short circuit!
        theta = 0.
    else:
        if xxbar != yybar:
            theta = math.atan(2. * xybar / (xxbar - yybar)) / 2.
        else:
            theta = numpy.sign(xybar) * math.pi / 4.0

        if theta * xybar > 0.:
            if theta < 0.:
                theta += math.pi / 2.0
            else:
                theta -= math.pi / 2.0

    ## NB: a dict should give us a bit more flexibility about arguments;
    ## however, all those here are ***REQUIRED***.
    return {
        "peak": peak,
        "flux": total,
        "xbar": xbar,
        "ybar": ybar,
        "semimajor": semimajor,
        "semiminor": semiminor,
        "theta": theta
    }


@guvectorize([(float32[:], int32[:], int32[:], int32[:], int32, int32, float32,
               float32, float32, float64, float64, float64[:], float64,
               float64[:], float64, float64, float32[:, :], float32[:, :])],
              ('(n), (m), (n), (n), (), (), (), (), (), (), (), (k), (), (m), ' +
               '(), (), (l, p) -> (l, p)'), nopython=True)
def moments_enhanced(island_data, chunkpos, posx, posy, min_width, no_pixels,
                     threshold, noise, maxi, fudge_max_pix_factor,
                     max_pix_variance_factor, beam, beamsize,
                     correlation_lengths,
                     clean_bias_error, frac_flux_cal_error, dummy,
                     computed_moments):
    """Calculate source properties using moments. Vectorized using the
    guvectorize decorator.

    Use the first moment of the distribution is the barycenter of an
    ellipse. The second moments are used to estimate the rotation angle
    and the length of the axes.

    Also, a positional 2D index local to the island is used such that every
    pixel value can be linked to a position relative to a corner of the island.

    Args:

        island_data (numpy.ndarray): Selected from the actual 2D image data,
                                     by taking pixels above the analysis
                                     threshold only, with its peak above the
                                     detection threshold. This selection
                                     results in a 1D ndarray (without a mask).

        chunkpos (numpy.ndarray): Index array of length 2 denoting the position
                                  of the top left corner of the rectangular
                                  slice encompassing the island relative to the
                                  top left corner of the image, which has pixel
                                  coordinates (0, 0), i.e. we need chunkpos
                                  to return to absolute pixel coordinates.

        posx (numpy.ndarray): Row indices of the pixels in island_data as taken
                              from the actual 2D images data
                              (rectangular slice).

        posy  (numpy.ndarray): Column indices of the pixels in island_data as
                               taken from the actual 2D images data (rectangular
                                slice).

        min_width (integer): The minimum width (in pixels) of the island. This
                             was derived by calculating the maximum width of the
                             island over x and y and then taking the minimum of
                             those two numbers.

        no_pixels (integer): The number of pixels that constitute the island.

        threshold(float): source parameters like the semimajor and semiminor
                          axes derived from moments can be underestimated if
                          one does not take account of the threshold that
                          was used to segment the source islands.

        noise(float): local noise, i.e. the standard deviation of the
                      background pixel values, at the position of the
                      peak pixel value of the island.

        maxi(float): peak pixel value from island.

        fudge_max_pix_factor(float): Correct for the underestimation of the peak
                                     by taking the maximum pixel value.

        max_pix_variance_factor (float): Take account of additional variance
                                        induced by the maximum pixel method,
                                        on top of the background noise.

        beam(numpy.ndarray): array from three floats: semimaj, semimin, theta.

        beamsize(float): The FWHM size of the clean beam

        correlation_lengths(numpy.ndarray): array from two floats describing the
                                    distance along the semi-major and semi-minor
                                    axes of the clean beam beyond which noise is
                                    assumed uncorrelated. Some background:
                                    Aperture synthesis imaging yields noise that
                                    is partially correlated over the entire
                                    image. This has a considerable effect on
                                    error estimates. We approximate this by
                                    considering all noise within the
                                    correlation length completely correlated
                                    and beyond that completely uncorrelated.

        clean_bias_error: Extra source of error copied from the
                          Condon (PASP 109, 166 (1997)) formulae

        frac_flux_cal_error: Extra source of error copied from the
                          Condon (PASP 109, 166 (1997)) formulae

        dummy (numpy.ndarray): Empty array with the same shape as
                               computed_moments needed because of a flau
                               in guvectorize: There is no other way to tell
                               guvectorize what the shape of the output array
                               will be. Therefore, we define an otherwise
                               redundant input array with the same shape as
                               the desired output array.

        computed_moments(numpy.ndarray): a (10, 2) array of floats containing
                                the computed moments, i.e.peak flux density,
                                total flux, x barycenter, y barycenter,
                                semimajor axis, semiminor axis, position angle
                                and the deconvolved counterparts of the latter
                                three quantities. This constitutes a total of
                                ten quantities and their corresponding errors.

    Returns:
        None (because of the guvectorize decorator), but computed_moments is
             filled with values.

    Raises:
        exceptions.ValueError: in case of NaN in input.

    """

    # Not every island has the same size. The number of columns of the array
    # containing all islands is equal to the maximum number of pxiels over
    # all islands. This containing array was created by numpy.empty, so better
    # dump the redundant elements that have undetermined values.
    island_data = island_data[:no_pixels]
    # Are we fitting a -ve or +ve Gaussian?
    if island_data.mean() >= 0:
        # The peak is always underestimated when you take the highest pixel.
        peak = maxi * fudge_max_pix_factor
    else:
        peak = island_data.min()

    ratio = threshold / peak
    total = island_data.sum()
    xbar, ybar, xxbar, yybar, xybar = 0, 0, 0, 0, 0

    for index in range(no_pixels):
        i = posx[index]
        j = posy[index]
        xbar += i * island_data[index]
        ybar += j * island_data[index]
        xxbar += i * i * island_data[index]
        yybar += j * j * island_data[index]
        xybar += i * j * island_data[index]

    xbar /= total
    ybar /= total
    xxbar /= total
    xxbar -= xbar ** 2
    yybar /= total
    yybar -= ybar ** 2
    xybar /= total
    xybar -= xbar * ybar

    working1 = (xxbar + yybar) / 2.0
    working2 = math.sqrt(((xxbar - yybar) / 2) ** 2 + xybar ** 2)
    smaj_tmp = (working1 + working2) * 2.0 * math.log(2.0)
    smin_tmp = (working1 - working2) * 2.0 * math.log(2.0)
    # ratio will be 0 for data that hasn't been selected according to a
    # threshold.
    if ratio != 0:
        # The corrections below for the semi-major and semi-minor axes are
        # to compensate for the underestimate of these quantities
        # due to the cutoff at the threshold.
        smaj_tmp /= (1.0 + math.log(ratio) * ratio / (1.0 - ratio))
        smin_tmp /= (1.0 + math.log(ratio) * ratio / (1.0 - ratio))

    # We need this width to determine Gaussian shape parameters in a meaningful
    # way.
    if min_width > 2:
        # The idea of the try except here is that, even though we require a
        # minimum width of the island, there may still be occasions where
        # working2 can be slightly higher than working1, perhaps due to rounding
        # errors.
        try:
            smaj = math.sqrt(smaj_tmp)
            smin = math.sqrt(smin_tmp)
            if smin == 0:
                # A semi-minor axis exactly zero gives all kinds of problems.
                # For instance wrt conversion to celestial coordinates.
                # This is a quick fix.
                smin = beamsize / (numpy.pi * smaj)

            if (numpy.isnan(xbar) or numpy.isnan(ybar) or
                    numpy.isnan(smaj) or numpy.isnan(smin)):
                raise ValueError("Unable to estimate Gauss shape")

            # Theta is not affected by the cut-off at the threshold (see Spreeuw 2010,
            # page 45).
            if abs(smaj - smin) < 0.01:
                # short circuit!
                theta = 0.
            else:
                if xxbar != yybar:
                    theta = math.atan(2. * xybar / (xxbar - yybar)) / 2.
                else:
                    theta = numpy.sign(xybar) * math.pi / 4.0

                if theta * xybar > 0.:
                    if theta < 0.:
                        theta += math.pi / 2.0
                    else:
                        theta -= math.pi / 2.0

        # In all cases where we hit a math domain error, we can use the clean
        # beam parameters to derive reasonable estimates for the Gaussian shape
        # parameters.
        except Exception:
            smaj = beam[0]
            smin = beam[1]
            theta = beam[2]
    else:
        # In the case that the island has insufficient width, we can also derive
        # reasonable estimates for Gaussian shape parameters using the clean
        # beam.
        smaj = beam[0]
        smin = beam[1]
        theta = beam[2]

    #  Equivalent of param["flux"] = (numpy.pi * param["peak"] *
    #  param["semimajor"] * param["semiminor"] / beamsize) from extract.py.
    flux = numpy.pi * peak * smaj * smin / beamsize

    # Update xbar and ybar with the position of the upper left corner of the
    # chunk.
    xbar += chunkpos[0]
    ybar += chunkpos[1]

    # There is no point in proceeding with processing an image if any
    # detected peak spectral brightness is below zero since that implies
    # that part of detectionthresholdmap is below zero.
    if peak < 0:
        raise ValueError

    """Provide reasonable error estimates from the moments"""

    # The formulae below should give some reasonable estimate of the
    # errors from moments, should always be higher than the errors from
    # Gauss fitting.
    theta_B = correlation_lengths[0]
    theta_b = correlation_lengths[1]

    # This is eq. 2.81 from Spreeuw's thesis.
    rho_sq = ((16. * smaj * smin /
               (numpy.log(2.) * theta_B * theta_b * noise ** 2))
              * ((peak - threshold) /
                 (numpy.log(peak) - numpy.log(threshold))) ** 2)

    rho = numpy.sqrt(rho_sq)
    denom = numpy.sqrt(2. * numpy.log(2.)) * rho

    # Again, like above for the Condon formulae, we set the
    # positional variances to twice the theoretical values.
    error_par_major = 2. * smaj / denom
    error_par_minor = 2. * smin / denom

    # When these errors are converted to RA and Dec,
    # calibration uncertainties will have to be added,
    # like in formulae 27 of the NVSS paper.
    errorx = numpy.sqrt((error_par_major * numpy.sin(theta)) ** 2
                        + (error_par_minor * numpy.cos(theta)) ** 2)
    errory = numpy.sqrt((error_par_major * numpy.cos(theta)) ** 2
                        + (error_par_minor * numpy.sin(theta)) ** 2)

    # Note that we report errors in HWHM axes instead of FWHM axes
    # so the errors are half the errors of formula 29 of the NVSS paper.
    errorsmaj = numpy.sqrt(2) * smaj / rho
    errorsmin = numpy.sqrt(2) * smin / rho

    if smaj > smin:
        errortheta = 2.0 * (smaj * smin / (smaj ** 2 - smin ** 2)) / rho
    else:
        errortheta = numpy.pi
    if errortheta > numpy.pi:
        errortheta = numpy.pi

    # The peak from "moments" is just the value of the maximum pixel
    # times a correction, fudge_max_pix, for the fact that the
    # centre of the Gaussian is not at the centre of the pixel.
    # This correction is performed in fitting.py. The maximum pixel
    # method introduces a peak dependent error corresponding to the last
    # term in the expression below for errorpeaksq.
    # To this, we add, in quadrature, the errors corresponding
    # to the first and last term of the rhs of equation 37 of the
    # NVSS paper. The middle term in that equation 37 is heuristically
    # replaced by noise**2 since the threshold should not affect
    # the error from the (corrected) maximum pixel method,
    # while it is part of the expression for rho_sq above.
    # errorpeaksq = ((frac_flux_cal_error * peak) ** 2 +
    #                clean_bias_error ** 2 + noise ** 2 +
    #                utils.maximum_pixel_method_variance(
    #                    beam[0], beam[1], beam[2]) * peak ** 2)
    errorpeaksq = ((frac_flux_cal_error * peak) ** 2 +
                   clean_bias_error ** 2 + noise ** 2 +
                   max_pix_variance_factor * peak ** 2)
    errorpeak = numpy.sqrt(errorpeaksq)

    help1 = (errorsmaj / smaj) ** 2
    help2 = (errorsmin / smin) ** 2
    help3 = theta_B * theta_b / (4. * smaj * smin)
    errorflux = flux * numpy.sqrt(
        errorpeaksq / peak ** 2 + help3 * (help1 + help2))

    """Deconvolve from the clean beam"""

    # If the fitted axes are smaller than the clean beam
    # (=restoring beam) axes, the axes and position angle
    # can be deconvolved from it.
    fmaj = 2. * smaj
    fmajerror = 2. * errorsmaj
    fmin = 2. * smin
    fminerror = 2. * errorsmin
    fpa = numpy.degrees(theta)
    fpaerror = numpy.degrees(errortheta)
    cmaj = 2. * beam[0]
    cmin = 2. * beam[1]
    cpa = numpy.degrees(beam[2])

    rmaj, rmin, rpa, ierr = deconv(fmaj, fmin, fpa, cmaj, cmin, cpa)
    # This parameter gives the number of components that could not be
    # deconvolved, IERR from deconf.f.
    deconv_imposs = ierr
    # Now, figure out the error bars.
    if rmaj > 0:
        # In this case the deconvolved position angle is defined.
        # For convenience we reset rpa to the interval [-90, 90].
        if rpa > 90:
            rpa = -numpy.mod(-rpa, 180.)
        theta_deconv = rpa

        # In the general case, where the restoring beam is elliptic,
        # calculating the error bars of the deconvolved position angle
        # is more complicated than in the NVSS case, where a circular
        # restoring beam was used.
        # In the NVSS case the error bars of the deconvolved angle are
        # equal to the fitted angle.
        rmaj1, rmin1, rpa1, ierr1 = deconv(
            fmaj, fmin, fpa + fpaerror, cmaj, cmin, cpa)
        if ierr1 < 2:
            if rpa1 > 90:
                rpa1 = -numpy.mod(-rpa1, 180.)
            rpaerror1 = numpy.abs(rpa1 - rpa)
            # An angle error can never be more than 90 degrees.
            if rpaerror1 > 90.:
                rpaerror1 = numpy.mod(-rpaerror1, 180.)
        else:
            rpaerror1 = numpy.nan
        rmaj2, rmin2, rpa2, ierr2 = deconv(
            fmaj, fmin, fpa - fpaerror, cmaj, cmin, cpa)
        if ierr2 < 2:
            if rpa2 > 90:
                rpa2 = -numpy.mod(-rpa2, 180.)
            rpaerror2 = numpy.abs(rpa2 - rpa)
            # An angle error can never be more than 90 degrees.
            if rpaerror2 > 90.:
                rpaerror2 = numpy.mod(-rpaerror2, 180.)
        else:
            rpaerror2 = numpy.nan
        if numpy.isnan(rpaerror1) or numpy.isnan(rpaerror2):
            theta_deconv_error = numpy.nansum(
                numpy.array([rpaerror1, rpaerror2]))
        else:
            theta_deconv_error = numpy.mean(
                numpy.array([rpaerror1, rpaerror2]))
        semimaj_deconv = rmaj / 2.
        rmaj3, rmin3, rpa3, ierr3 = deconv(
            fmaj + fmajerror, fmin, fpa, cmaj, cmin, cpa)
        # If rmaj>0, then rmaj3 should also be > 0,
        # if I am not mistaken, see the formulas at
        # the end of ch.2 of Spreeuw's Ph.D. thesis.
        if fmaj - fmajerror > fmin:
            rmaj4, rmin4, rpa4, ierr4 = deconv(
                fmaj - fmajerror, fmin, fpa, cmaj, cmin, cpa)
            if rmaj4 > 0:
                semimaj_deconv_error = numpy.mean(numpy.array(
                    [numpy.abs(rmaj3 - rmaj), numpy.abs(rmaj - rmaj4)]))
            else:
                semimaj_deconv_error = numpy.abs(rmaj3 - rmaj)
        else:
            rmin4, rmaj4, rpa4, ierr4 = deconv(
                fmin, fmaj - fmajerror, fpa, cmaj, cmin, cpa)
            if rmaj4 > 0:
                semimaj_deconv_error = numpy.mean(numpy.array(
                    [numpy.abs(rmaj3 - rmaj), numpy.abs(rmaj - rmaj4)]))
            else:
                semimaj_deconv_error = numpy.abs(rmaj3 - rmaj)
        if rmin > 0:
            semimin_deconv = rmin / 2.
            if fmin + fminerror < fmaj:
                rmaj5, rmin5, rpa5, ierr5 = deconv(
                    fmaj, fmin + fminerror, fpa, cmaj, cmin, cpa)
            else:
                rmin5, rmaj5, rpa5, ierr5 = deconv(
                    fmin + fminerror, fmaj, fpa, cmaj, cmin, cpa)
            # If rmin > 0, then rmin5 should also be > 0,
            # if I am not mistaken, see the formulas at
            # the end of ch.2 of Spreeuw's Ph.D. thesis.
            rmaj6, rmin6, rpa6, ierr6 = deconv(
                fmaj, fmin - fminerror, fpa, cmaj, cmin, cpa)
            if rmin6 > 0:
                semimin_deconv_error = numpy.mean(numpy.array(
                    [numpy.abs(rmin6 - rmin), numpy.abs(rmin5 - rmin)]))
            else:
                semimin_deconv_error = numpy.abs(rmin5 - rmin)
        else:
            semimin_deconv = numpy.nan
            semimin_deconv_error = numpy.nan
    else:
        semimaj_deconv = numpy.nan
        semimaj_deconv_error = numpy.nan
        semimin_deconv = numpy.nan
        semimin_deconv_error = numpy.nan
        theta_deconv = numpy.nan
        theta_deconv_error= numpy.nan

    computed_moments[0, :] = numpy.array([peak, flux, xbar, ybar, smaj,
                                          smin, theta, semimaj_deconv,
                                          semimin_deconv, theta_deconv])
    computed_moments[1, :] = numpy.array([errorpeak, errorflux, errorx,
                                          errory, errorsmaj, errorsmin,
                                          errortheta, semimaj_deconv_error,
                                          semimin_deconv_error,
                                          theta_deconv_error])


def fitgaussian(pixels, params, fixed=None, max_nfev=None, bounds={}):
    """Calculate source positional values by fitting a 2D Gaussian

    :Args:
        pixels (numpy.ma.MaskedArray): Pixel values (with bad pixels masked)

        params (dict): initial fit parameters (possibly estimated
            using the moments() function, above)

    :Kwargs:
        fixed (dict): parameters & their values to be kept frozen (ie, not
            fitted)

        max_nfev (int): maximum number of calls to the error function

        bounds (dict): can be a dict such as the extract.ParamSet().bounds
            attribute generated by the extract.ParamSet().compute_bounds method,
            but any dict with keys from FIT_PARAMS and
            (lower_bound, upper_bound, bool) tuples as values will do. The
            boolean argument accommodates for loosening a bound when a fit
            becomes unfeasible because of the bound, see the
            scipy.optimize.Bounds documentation and source code for background.
            lower_bound and upper_bound are floats.

    Returns:
        dict: peak, total, x barycenter, y barycenter, semimajor,
            semiminor, theta (radians)

    Raises:
        exceptions.ValueError: In case of a bad fit.

    Perform a least squares fit to an elliptical Gaussian.

    If a dict called fixed is passed in, then parameters specified within the
    dict with the same names as fit_params (below) will be "locked" in the
    fitting process.
    """
    fixed = fixed or {}

    def wipe_out_fixed(some_dict):
        wiped_out = map(lambda key: (key, some_dict[key]),
                        filter(lambda key: key not in fixed.keys(),
                        iter(FIT_PARAMS)))
        return dict(wiped_out)

    # Collect necessary values from parameter dict; only those which aren't
    # fixed.
    initial = []
    # To avoid fits from derailing, it helps to set some reasonable bounds on
    # the fitting results.
    for param in FIT_PARAMS:
        if param not in fixed:
            if hasattr(params[param], "value"):
                initial.append(params[param].value)
            else:
                initial.append(params[param])

    def residuals(params):
        """Error function to be used in chi-squared fitting

        Args:
            params(numpy.ndarray): fitting parameters

        Returns:
            1d-array of difference between estimated Gaussian function
            and the actual pixels. (pixel_resids is a 2d-array, but the
            .compressed() makes it 1d.)
        """
        paramlist = list(params)
        gaussian_args = []
        for param in FIT_PARAMS:
            if param in fixed:
                gaussian_args.append(fixed[param])
            else:
                gaussian_args.append(paramlist.pop(0))

        # gaussian() returns a function which takes arguments x, y and returns
        # a Gaussian with parameters gaussian_args evaluated at that point.
        g = gaussian(*gaussian_args)

        # The .compressed() below is essential so the Gaussian fit will not
        # take account of the masked values (=below threshold) at the edges
        # and corners of pixels (=(masked) array, so rectangular).
        pixel_resids = numpy.ma.MaskedArray(
            data=numpy.fromfunction(g, pixels.shape) - pixels,
            mask=pixels.mask)

        return pixel_resids.compressed()

    def jacobian_values(params):
        """The Jacobian of an anisotropic 2D Gaussian at the pixel positions.

        Args:
            params(numpy.ndarray): fitting parameters

        Returns:
             2d-array with values of the partial derivatives of the
             2d anisotropic Gaussian along its six parameters. These
             values are evaluated across the unmasked pixel positions of
             the island that constitutes the source. The number of rows
             equals the number of pixels of the flattened unmasked island.
             The number of columns equals the number of partial
             derivatives of the Gaussian (=6). For fixed Gaussian
             parameters the Jacobian component is obviously zero, so that
             results in a column of only zeroes.

        """
        paramlist = list(params)
        gaussian_args = []
        for param in FIT_PARAMS:
            if param in fixed:
                gaussian_args.append(fixed[param])
            else:
                gaussian_args.append(paramlist.pop(0))

        # jac is a list of six functions corresponding to the six partial
        # derivatives of the 2D anisotropic Gaussian profile.
        jac = jac_gaussian(gaussian_args)

        # From the six functions in jac, wipe out the ones that
        # correspond to fixed parameters, must be in sync with initial.
        jac_filtered = wipe_out_fixed(jac)

        jac_values = [numpy.ma.MaskedArray(
            data=numpy.fromfunction(jac_filtered[key], pixels.shape),
            mask=pixels.mask).compressed() for key in jac_filtered]

        return numpy.array(jac_values).T

    # maxfev=0, the default, corresponds to 200*(N+1) (NB, not 100*(N+1) as
    # the scipy docs state!) function evaluations, where N is the number of
    # parameters in the solution.
    # Convergence tolerances xtol and ftol established by experiment on images
    # from Paul Hancock's simulations.
    # April 2024 update.
    # scipy.optimize.leastsq was replaced by scipy.optimize.least_squares,
    # because it offers more fitting options. max_nfev instead of maxfev is one
    # of its keyword arguments, with a default value of None instead of 0.

    if bounds:
        # Wipe out fixed parameters, keep synced with initial.
        bounds_filtered = wipe_out_fixed(bounds)
        # bounds and bounds_filtered are dicts of 3-tuples with the third
        # element of those tuples a Boolean value, which can be passed to a
        # scipy.optimize.Bounds instance to allow for loosening of the bounds
        # when these bounds turn out infeasible in the fitting process. However,
        # The 3-tuples need to extracted into separate array_like objects to be
        # passed on as lb, ub and keep_feasible keyword arguments to the Bounds
        # instance.
        bounds_filtered_values = [[bounds_filtered[x][index] for x in
                                   bounds_filtered] for index in range(3)]
        lb = bounds_filtered_values[0]
        ub = bounds_filtered_values[1]
        keep_feasible = bounds_filtered_values[2]
        applied_bounds = scipy.optimize.Bounds(lb=lb, ub=ub,
                                               keep_feasible=keep_feasible)
    else:
        applied_bounds = (-numpy.inf, numpy.inf)

    Fitting_results = scipy.optimize.least_squares(
        residuals, initial, jac=jacobian_values,
        bounds=applied_bounds, method="dogbox",
        max_nfev=max_nfev, xtol=1e-4, ftol=1e-4
    )

    soln = Fitting_results["x"]
    success = Fitting_results["success"]

    if success > 4:
        raise ValueError("leastsq returned %d; bailing out" % (success,))

    # soln contains only the variable parameters; we need to merge the
    # contents of fixed into the soln list.
    # leastsq() returns either a numpy.float64 (if fitting a single value) or
    # a numpy.ndarray (if fitting multiple values); we need to turn that into
    # a list for the merger.
    try:
        # If a ndarray (or other iterable)
        soln = list(soln)
    except TypeError:
        soln = [soln]
    results = fixed.copy()
    for param in FIT_PARAMS:
        if param not in results:
            results[param] = soln.pop(0)

    if results['semiminor'] > results['semimajor']:
        # Swapped axis order is a perfectly valid fit, but inconvenient for
        # the rest of our codebase.
        results['semimajor'], results['semiminor'] = results['semiminor'], \
                                                     results['semimajor']
        results['theta'] += numpy.pi / 2

    # Negative axes are a valid fit, since they are squared in the definition
    # of the Gaussian.
    results['semimajor'] = abs(results['semimajor'])
    results['semiminor'] = abs(results['semiminor'])

    return results


def goodness_of_fit(masked_residuals, noise, correlation_lengths):
    """
    Calculates the goodness-of-fit values, `chisq` and `reduced_chisq`.

    .. Warning::
        We do not use the `standard chi-squared
        formula <https://en.wikipedia.org/wiki/Goodness_of_fit#Regression_analysis>`_
        for calculating these goodness-of-fit
        values, and should probably rename them in the next release.
        See below for details.


    These goodness-of-fit values are related to, but not quite the same as
    reduced chi-squared.
    Strictly speaking the reduced chi-squared is statistically
    invalid for a Gaussian model from the outset
    (see `arxiv:1012.3754 <http://arxiv.org/abs/1012.3754>`_).
    We attempt to provide a resolution-independent estimate of goodness-of-fit
    ('reduced chi-squared'), by using the same 'independent pixels' correction
    as employed when estimating RMS levels, to normalize the chi-squared value.
    However, as applied to the standard formula this will sometimes
    imply that we are fitting a fractional number of datapoints less than 1!
    As a result, it doesn't really make sense to try and apply the
    'degrees-of-freedom' correction, as this would likely result in a
    negative ``reduced_chisq`` value.
    (And besides, the 'degrees of freedom' concept is invalid for non-linear
    models.) Finally, note that when called from
    :func:`.source_profile_and_errors`, the noise-estimate at the peak-pixel
    is supplied, so will typically over-estimate the noise and
    hence under-estimate the chi-squared values.

    Args:
        masked_residuals(numpy.ma.MaskedArray): The pixel-residuals from the fit
        noise (float): An estimate of the noise level. Could also be set to
            a masked numpy array matching the data, for per-pixel noise
            estimates.

        correlation_lengths(tuple): Tuple of two floats describing the distance along the semimajor
                                    and semiminor axes of the clean beam beyond which noise
                                    is assumed uncorrelated. Some background: Aperture synthesis imaging
                                    yields noise that is partially correlated
                                    over the entire image. This has a considerable effect on error
                                    estimates. We approximate this by considering all noise within the
                                    correlation length completely correlated and beyond that completely
                                    uncorrelated.

    Returns:
        tuple: chisq, reduced_chisq

    """
    gauss_resid_normed = (masked_residuals / noise).compressed()
    chisq = numpy.sum(gauss_resid_normed * gauss_resid_normed)
    n_fitted_pix = len(masked_residuals.compressed().ravel())
    n_indep_pix = indep_pixels(n_fitted_pix, correlation_lengths)
    reduced_chisq = chisq / n_indep_pix
    return chisq, reduced_chisq



import logging
import os.path
from collections import namedtuple

import astropy.io.fits as pyfits
from casacore.images import image as casacore_image
from casacore.tables import table as casacore_table

from sourcefinder.accessors.aartfaaccasaimage import AartfaacCasaImage
from sourcefinder.accessors.amicasaimage import AmiCasaImage
from sourcefinder.accessors.fitsimage import FitsImage
from sourcefinder.accessors.kat7casaimage import Kat7CasaImage
from sourcefinder.accessors.lofarcasaimage import LofarCasaImage
from sourcefinder.accessors.lofarfitsimage import LofarFitsImage
from sourcefinder.accessors.lofarhdf5image import LofarHdf5Image

logger = logging.getLogger(__name__)

# files that should be contained by a casa table
casafiles = ("table.dat", "table.f0", "table.f0_TSM0", "table.info",
             "table.lock")

# We will take the first accessor for which the test returns True.
FitsTest = namedtuple('FitsTest', ['accessor', 'test'])
fits_type_mapping = [
    FitsTest(
        accessor=LofarFitsImage,
        test=lambda hdr: 'TELESCOP' in hdr and 'ANTENNA' in hdr and hdr.get(
            'TELESCOP') == "LOFAR"
    )
]

casa_telescope_keyword_mapping = {
    'LOFAR': LofarCasaImage,
    'KAT-7': Kat7CasaImage,
    'AARTFAAC': AartfaacCasaImage,
    'AMI-LA': AmiCasaImage,
}


def isfits(filename):
    """returns True if filename is a fits file"""
    if not os.path.isfile(filename):
        return False
    if filename[-4:].lower() != 'fits':
        return False
    try:
        with pyfits.open(filename):
            pass
    except IOError:
        return False
    return True


def iscasa(filename):
    """returns True if filename is a lofar casa directory"""
    if not os.path.isdir(filename):
        return False
    for file_ in casafiles:
        casafile = os.path.join(filename, file_)
        if not os.path.isfile(casafile):
            logger.debug("%s doesn't contain %s" % (filename, file_))
            return False
    try:
        table = casacore_table(filename, ack=False)
        table.close()
    except RuntimeError as e:
        logger.debug("directory looks casacore, but cannot open: %s" % str(e))
        return False
    return True


def islofarhdf5(filename):
    """returns True if filename is a hdf5 container"""
    if not os.path.isfile(filename):
        return False
    if filename[-2:].lower() != 'h5':
        return False
    try:
        casacore_image(filename)
    except RuntimeError:
        return False
    return True


def fits_detect(filename):
    """
    Detect which telescope produced FITS data, return corresponding accessor.

    Checks for known FITS image types where we expect additional metadata.
    If the telescope is unknown we default to a regular FitsImage.
    """
    with pyfits.open(filename) as hdulist:
        hdr = hdulist[0].header
    for fits_test in fits_type_mapping:
        if fits_test.test(hdr):
            return fits_test.accessor
    return FitsImage


def casa_detect(filename):
    """
    Detect which telescope produced CASA data, return corresponding accessor.

    Checks for known CASA table types where we expect additional metadata.
    If the telescope is unknown we return nothing.
    """
    table = casacore_table(filename, ack=False)
    telescope = table.getkeyword('coords')['telescope']
    return casa_telescope_keyword_mapping.get(telescope, None)


def detect(filename):
    """returns the accessor class that should be used to process filename"""
    if isfits(filename):
        return fits_detect(filename)
    elif iscasa(filename):
        return casa_detect(filename)
    elif islofarhdf5(filename):
        return LofarHdf5Image
    else:
        raise OSError("unsupported format: %s" % filename)



"""
Source Extraction Helpers.

These are used in conjunction with image.ImageData.
"""

from sourcefinder.deconv import deconv
from sourcefinder.utility import coordinates
from sourcefinder.utility.uncertain import Uncertain
from .gaussian import gaussian
from . import fitting
from . import utils
import logging
from collections.abc import MutableMapping
from numba import guvectorize, float64, float32, int32
import numpy
numpy.seterr(divide="raise", invalid="raise")

try:
    import ndimage
except ImportError:
    from scipy import ndimage

logger = logging.getLogger(__name__)

# This is used as a dummy value, -BIGNUM values will be always be masked.
# As such, it should be larger than the expected range of real values,
# since we will get negative values representing real data after
# background subtraction, etc.
BIGNUM = 99999.0


class Island(object):
    """
    The source extraction process forms islands, which it then fits.
    Each island needs to know its position in the image (ie, x, y pixel
    value at one corner), the threshold above which it is detected
    (analysis_threshold by default, but will increase if the island is
    the result of deblending), and a data array.

    The island should provide a means of deblending: splitting itself
    apart and returning multiple sub-islands, if necessary.
    """

    def __init__(self, data, rms, chunk, analysis_threshold, detection_map,
                 beam, deblend_nthresh, deblend_mincont, structuring_element,
                 rms_orig=None, flux_orig=None, subthrrange=None
                 ):

        # deblend_nthresh is the number of subthresholds used when deblending.
        self.deblend_nthresh = deblend_nthresh
        # If we deblend too far, we hit the recursion limit. And it's slow.
        if self.deblend_nthresh > 300:
            logger.warning("Limiting to 300 deblending subtresholds")
            self.deblend_nthresh = 300
        else:
            logger.debug("Using %d subthresholds", deblend_nthresh)

        # Deblended components of this island must contain at least
        # deblend_mincont times the total flux of the original to be regarded
        # as significant.
        self.deblend_mincont = deblend_mincont

        # The structuring element defines connectivity between pixels.
        self.structuring_element = structuring_element

        # NB we have set all unused data to -(lots) before passing it to
        # Island().
        mask = numpy.where(data > -BIGNUM / 10.0, 0, 1)
        self.data = numpy.ma.array(data, mask=mask)
        self.rms = rms
        self.chunk = chunk
        self.analysis_threshold = analysis_threshold
        self.detection_map = detection_map
        self.beam = beam
        self.max_pos = ndimage.maximum_position(self.data.filled(fill_value=0))
        self.position = (self.chunk[0].start, self.chunk[1].start)
        if not isinstance(rms_orig, numpy.ndarray):
            self.rms_orig = self.rms
        else:
            self.rms_orig = rms_orig
        # The idea here is to retain the flux of the original, unblended
        # island. That flux is used as a criterion for deblending.
        if not isinstance(flux_orig, float):
            self.flux_orig = self.data.sum()
        else:
            self.flux_orig = flux_orig
        if isinstance(subthrrange, numpy.ndarray):
            self.subthrrange = subthrrange
        else:
            self.subthrrange = utils.generate_subthresholds(
                self.data.min(), self.data.max(), self.deblend_nthresh
            )

    def deblend(self, niter=0):
        """Return a decomposed numpy array of all the subislands.

        Iterate up through subthresholds, looking for our island
        splitting into two. If it does, start again, with two or more
        separate islands.
        """

        logger.debug("Deblending source")
        for level in self.subthrrange[niter:]:

            # The idea is to retain the parent island when no significant
            # subislands are found and jump to the next subthreshold
            # using niter.
            # Deblending is started at a level higher than the lowest
            # pixel value in the island.
            # Deblending at the level of the lowest pixel value will
            # likely not yield anything, because the island was formed at
            # threshold just below that.
            # So that is why we use niter+1 (>=1) instead of niter (>=0).

            if level > self.data.max():
                # level is above the highest pixel value...
                # Return the current island.
                break
            clipped_data = numpy.where(
                self.data.filled(fill_value=0) >= level, 1, 0)
            labels, number = ndimage.label(clipped_data, self.structuring_element)

            # If we have more than one island, then we need to make subislands.
            if number > 1:
                subislands = []
                label = 0
                for chunk in ndimage.find_objects(labels):
                    label += 1
                    newdata = numpy.where(
                        labels == label,
                        self.data.filled(fill_value=-BIGNUM), -BIGNUM
                    )
                    # NB: In class Island(object), rms * analysis_threshold
                    # is taken as the threshold for the bottom of the island.
                    # Everything below that level is masked.
                    # For subislands, this product should be equal to level
                    # and flat, i.e., horizontal.
                    # We can achieve this by setting rms=level*ones and
                    # analysis_threshold=1.
                    island = Island(
                        newdata[chunk],
                        (numpy.ones(self.data[chunk].shape) * level),
                        (
                            slice(self.chunk[0].start + chunk[0].start,
                                  self.chunk[0].start + chunk[0].stop),
                            slice(self.chunk[1].start + chunk[1].start,
                                  self.chunk[1].start + chunk[1].stop)
                        ),
                        1,
                        self.detection_map[chunk],
                        self.beam,
                        self.deblend_nthresh,
                        self.deblend_mincont,
                        self.structuring_element,
                        self.rms_orig[chunk[0].start:chunk[0].stop,
                                      chunk[1].start:chunk[1].stop],
                        self.flux_orig,
                        self.subthrrange
                    )

                    subislands.append(island)
                # This line should filter out any subisland with insufficient
                # flux, in about the same way as SExtractor.
                # Sufficient means: the flux of the branch above the
                # subthreshold (=level) must exceed some user given fraction
                # of the composite object, i.e., the original island.
                subislands = [isl for isl in subislands if (isl.data - numpy.ma.array(
                    numpy.ones(isl.data.shape) * level,
                    mask=isl.data.mask)).sum() > self.deblend_mincont *
                              self.flux_orig]
                # Discard subislands below detection threshold
                subislands = [isl for isl in subislands if (isl.data - isl.detection_map).max() >= 0]
                numbersignifsub = len(subislands)
                # Proceed with the previous island, but make sure the next
                # subthreshold is higher than the present one.
                # Or we would end up in an infinite loop...
                if numbersignifsub > 1:
                    if niter + 1 < self.deblend_nthresh:
                        # Apparently, the map command always results in
                        # nested lists.
                        return list(utils.flatten([island.deblend(niter=niter + 1) for island in subislands]))
                    else:
                        return subislands
                elif numbersignifsub == 1 and niter + 1 < self.deblend_nthresh:
                    return Island.deblend(self, niter=niter + 1)
                else:
                    # In this case we have numbersignifsub == 0 or
                    # (1 and reached the highest subthreshold level).
                    # Pull out of deblending loop, return current island.
                    break
        # We've not found any subislands: just return this island.
        return self

    def threshold(self):
        """Threshold"""
        return self.noise() * self.analysis_threshold

    def noise(self):
        """Noise at maximum position"""
        return self.rms[self.max_pos]

    def sig(self):
        """Deviation"""
        return (self.data / self.rms_orig).max()

    def fit(self, fudge_max_pix_factor, max_pix_variance_factor, beamsize,
            correlation_lengths, fixed=None):
        """Fit the position"""
        try:
            measurement, gauss_island, gauss_residual = \
                source_profile_and_errors(self.data, self.threshold(),
                                          self.noise(), self.beam,
                                          fudge_max_pix_factor,
                                          max_pix_variance_factor, beamsize,
                                          correlation_lengths, fixed=fixed)
        except ValueError:
            # Fitting failed
            logger.error("Moments & Gaussian fitting failed at %s" % (
                str(self.position)))
            return None
        measurement["xbar"] += self.position[0]
        measurement["ybar"] += self.position[1]
        measurement.sig = self.sig()
        return measurement, gauss_island, gauss_residual


class ParamSet(MutableMapping):
    """
    All the source fitting methods should go to produce a ParamSet, which
    gives all the information necessary to make a Detection.
    """

    def __init__(self, clean_bias=0.0, clean_bias_error=0.0,
                 frac_flux_cal_error=0.0, alpha_maj1=2.5, alpha_min1=0.5,
                 alpha_maj2=0.5, alpha_min2=2.5, alpha_maj3=1.5,
                 alpha_min3=1.5):

        self.clean_bias = clean_bias
        self.clean_bias_error = clean_bias_error
        self.frac_flux_cal_error = frac_flux_cal_error
        self.alpha_maj1 = alpha_maj1
        self.alpha_min1 = alpha_min1
        self.alpha_maj2 = alpha_maj2
        self.alpha_min2 = alpha_min2
        self.alpha_maj3 = alpha_maj3
        self.alpha_min3 = alpha_min3

        self.measurements = {
            'peak': Uncertain(),
            'flux': Uncertain(),
            'xbar': Uncertain(),
            'ybar': Uncertain(),
            'semimajor': Uncertain(),
            'semiminor': Uncertain(),
            'theta': Uncertain(),
            'semimaj_deconv': Uncertain(),
            'semimin_deconv': Uncertain(),
            'theta_deconv': Uncertain()
        }
        # This parameter gives the number of components that could not be
        # deconvolved, IERR from deconf.f.
        self.deconv_imposs = 2

        # These flags are used to indicate where the values stored in this
        # parameterset have come from: we set them to True if & when moments
        # and/or Gaussian fitting succeeds.
        self.moments = False
        # Gaussian fits are preferable performed with bounds for better
        # stability, but to establish those bounds in a meaningful manner
        # moments estimation is required, i.e. self.moments = True.
        self.bounds = {}
        self.gaussian = False

        # More metadata about the fit: only valid for Gaussian fits:
        self.chisq = None
        self.reduced_chisq = None

    def __getitem__(self, item):
        return self.measurements[item]

    def __setitem__(self, item, value):
        if item in self.measurements:
            if isinstance(value, Uncertain):
                self.measurements[item] = value
            else:
                self.measurements[item].value = value
        else:
            raise AttributeError("Invalid parameter")

    def __delitem__(self, key):
        del self.measurements[key]

    def __iter__(self):
        return iter(self.measurements)

    def __len__(self):
        return len(self.measurements)

    def keys(self):
        """ """
        return list(self.measurements.keys())

    def compute_bounds(self, data_shape):
        """ Calculate bounds for 'safer' Gauss fitting, i.e. a smaller chance
        on runaway solutions. The bounds are largely based on moments
        estimation, so it only makes sense to impose bounds if moments
        estimation was successful.

        Args:
            data_shape: (int32, int32) or (int64, int64) tuple describing the
                        shape of the rectangular area (slice) encompassing the
                        island used for the fit. Used to set bounds for the
                        position of the source in the fitting process.

        Creates dict of (float, float, bool) tuples, i.e. for a maximum of six
        Gaussian fit parameters a (lower_bound, upper_bound, bool) tuple. The
        boolean entry is used to loosen a bound when a fit becomes unfeasible,
        see the documentation on scipy.optimize.Bounds.
        """
        if self.moments:
            if hasattr(self["peak"], "value"):
                self.bounds["peak"] = (0.5 * self["peak"].value,
                                       1.5 * self["peak"].value, False)
            else:
                self.bounds["peak"] = (0.5 * self["peak"], 1.5 * self["peak"],
                                       False)
            # Accommodate for negative heights. The "bounds" argument in
            # scipy.optimize.least-squares demands that the lower bound is
            # smaller than the upper bound and with the (0.5, 1.5) fractions
            # this does not work out well for negative heights. Background:
            # only to be expected in images from visibilities from polarization
            # products other than Stokes I.
            if self.bounds["peak"][0] > self.bounds["peak"][1]:
                true_upper = self.bounds["peak"][0]
                self.bounds["peak"][0] = self.bounds["peak"][1]
                self.bounds["peak"][1] = true_upper

            if hasattr(self["semimajor"], "value"):
                self.bounds["semimajor"] = (0.5 * self["semimajor"].value,
                                            1.5 * self["semimajor"].value,
                                            False)
            else:
                self.bounds["semimajor"] = (0.5 * self["semimajor"],
                                            1.5 * self["semimajor"], False)

            if hasattr(self["semiminor"], "value"):
                self.bounds["semiminor"] = (0.5 * self["semiminor"].value,
                                            1.5 * self["semiminor"].value,
                                            False)
            else:
                self.bounds["semiminor"] = (0.5 * self["semiminor"],
                                            1.5 * self["semiminor"], False)

            self.bounds["xbar"] = (0., data_shape[0], False)
            self.bounds["ybar"] = (0., data_shape[1], False)
            # The upper bound for theta is a bit odd, one would expect
            # numpy.pi/2 here, but that will yield imperfect fits in
            # cases where the axes are aligned with the coordinate axes,
            # which, in turn, breaks the AxesSwapGaussTest.testFitHeight
            # and AxesSwapGaussTest.testFitSize unit tests. Thus, some
            # margin needs to be applied. 1.05 was chosen here.
            # keep_feasibly is True here to accommodate for a fitting process
            # where the margin is exceeded.
            self.bounds["theta"] = (-numpy.pi/2, numpy.pi, True)

        return self

    def calculate_errors(self, noise, max_pix_variance_factor, correlation_lengths, threshold):
        """Calculate positional errors

        Uses _condon_formulae() if this object is based on a Gaussian fit,
        _error_bars_from_moments() if it's based on moments.
        """

        if self.gaussian:
            return self._condon_formulae(noise, correlation_lengths)
        elif self.moments:
            if not threshold:
                threshold = 0
            return self._error_bars_from_moments(noise, max_pix_variance_factor, correlation_lengths,
                                                 threshold)
        else:
            return False

    def _condon_formulae(self, noise, correlation_lengths):
        """Returns the errors on parameters from Gaussian fits according to
        the Condon (PASP 109, 166 (1997)) formulae.

        These formulae are not perfect, but we'll use them for the
        time being.  (See Refregier and Brown (astro-ph/9803279v1) for
        a more rigorous approach.) It also returns the corrected peak.
        The peak is corrected for the overestimate due to the local
        noise gradient.
        """
        peak = self['peak'].value
        flux = self['flux'].value
        smaj = self['semimajor'].value
        smin = self['semiminor'].value
        theta = self['theta'].value

        theta_B, theta_b = correlation_lengths

        rho_sq1 = ((smaj * smin / (theta_B * theta_b)) *
                   (1. + (theta_B / (2. * smaj)) ** 2) ** self.alpha_maj1 *
                   (1. + (theta_b / (2. * smin)) ** 2) ** self.alpha_min1 *
                   (peak / noise) ** 2)
        rho_sq2 = ((smaj * smin / (theta_B * theta_b)) *
                   (1. + (theta_B / (2. * smaj)) ** 2) ** self.alpha_maj2 *
                   (1. + (theta_b / (2. * smin)) ** 2) ** self.alpha_min2 *
                   (peak / noise) ** 2)
        rho_sq3 = ((smaj * smin / (theta_B * theta_b)) *
                   (1. + (theta_B / (2. * smaj)) ** 2) ** self.alpha_maj3 *
                   (1. + (theta_b / (2. * smin)) ** 2) ** self.alpha_min3 *
                   (peak / noise) ** 2)

        rho1 = numpy.sqrt(rho_sq1)
        rho2 = numpy.sqrt(rho_sq2)
        rho3 = numpy.sqrt(rho_sq3)

        denom1 = numpy.sqrt(2. * numpy.log(2.)) * rho1
        denom2 = numpy.sqrt(2. * numpy.log(2.)) * rho2

        # Here you get the errors parallel to the fitted semi-major and
        # semi-minor axes as taken from the NVSS paper (Condon et al. 1998,
        # AJ, 115, 1693), formula 25.
        # Those variances are twice the theoreticals, so the errors in
        # position are sqrt(2) as large as one would get from formula 21
        # of the Condon (1997) paper.
        error_par_major = 2. * smaj / denom1
        error_par_minor = 2. * smin / denom2

        # When these errors are converted to RA and Dec,
        # calibration uncertainties will have to be added,
        # like in formulae 27 of the NVSS paper.
        errorx = numpy.sqrt((error_par_major * numpy.sin(theta)) ** 2 +
                            (error_par_minor * numpy.cos(theta)) ** 2)
        errory = numpy.sqrt((error_par_major * numpy.cos(theta)) ** 2 +
                            (error_par_minor * numpy.sin(theta)) ** 2)

        # Note that we report errors in HWHM axes instead of FWHM axes
        # so the errors are half the errors of formula 29 of the NVSS paper.
        errorsmaj = numpy.sqrt(2) * smaj / rho1
        errorsmin = numpy.sqrt(2) * smin / rho2

        if smaj > smin:
            errortheta = 2.0 * (smaj * smin / (smaj ** 2 - smin ** 2)) / rho2
        else:
            errortheta = numpy.pi
        if errortheta > numpy.pi:
            errortheta = numpy.pi

        peak += -noise ** 2 / peak + self.clean_bias

        errorpeaksq = ((self.frac_flux_cal_error * peak) ** 2 +
                       self.clean_bias_error ** 2 +
                       2. * peak ** 2 / rho_sq3)

        errorpeak = numpy.sqrt(errorpeaksq)

        help1 = (errorsmaj / smaj) ** 2
        help2 = (errorsmin / smin) ** 2
        help3 = theta_B * theta_b / (4. * smaj * smin)
        errorflux = numpy.abs(flux) * numpy.sqrt(
            errorpeaksq / peak ** 2 + help3 * (help1 + help2))

        self['peak'] = Uncertain(peak, errorpeak)
        self['flux'].error = errorflux
        self['xbar'].error = errorx
        self['ybar'].error = errory
        self['semimajor'].error = errorsmaj
        self['semiminor'].error = errorsmin
        self['theta'].error = errortheta

        return self

    def _error_bars_from_moments(self, noise, max_pix_variance_factor, correlation_lengths,
                                 threshold):
        """Provide reasonable error estimates from the moments"""

        # The formulae below should give some reasonable estimate of the
        # errors from moments, should always be higher than the errors from
        # Gauss fitting.
        peak = self['peak'].value
        flux = self['flux'].value
        smaj = self['semimajor'].value
        smin = self['semiminor'].value
        theta = self['theta'].value

        # There is no point in proceeding with processing an image if any
        # detected peak spectral brightness is below zero since that implies
        # that part of detectionthresholdmap is below zero.
        if peak < 0:
            raise ValueError((f"Peak from moments = {peak:.2e}, something is "
                              "possibly wrong with detectionthresholdmap, "
                              "since a negative peak from moments analysis "
                              "should not occur."))

        clean_bias_error = self.clean_bias_error
        frac_flux_cal_error = self.frac_flux_cal_error
        theta_B, theta_b = correlation_lengths

        # This is eq. 2.81 from Spreeuw's thesis.
        rho_sq = ((16. * smaj * smin /
                   (numpy.log(2.) * theta_B * theta_b * noise ** 2))
                  * ((peak - threshold) /
                     (numpy.log(peak) - numpy.log(threshold))) ** 2)

        rho = numpy.sqrt(rho_sq)
        denom = numpy.sqrt(2. * numpy.log(2.)) * rho

        # Again, like above for the Condon formulae, we set the
        # positional variances to twice the theoretical values.
        error_par_major = 2. * smaj / denom
        error_par_minor = 2. * smin / denom

        # When these errors are converted to RA and Dec,
        # calibration uncertainties will have to be added,
        # like in formulae 27 of the NVSS paper.
        errorx = numpy.sqrt((error_par_major * numpy.sin(theta)) ** 2
                            + (error_par_minor * numpy.cos(theta)) ** 2)
        errory = numpy.sqrt((error_par_major * numpy.cos(theta)) ** 2
                            + (error_par_minor * numpy.sin(theta)) ** 2)

        # Note that we report errors in HWHM axes instead of FWHM axes
        # so the errors are half the errors of formula 29 of the NVSS paper.
        errorsmaj = numpy.sqrt(2) * smaj / rho
        errorsmin = numpy.sqrt(2) * smin / rho

        if smaj > smin:
            errortheta = 2.0 * (smaj * smin / (smaj ** 2 - smin ** 2)) / rho
        else:
            errortheta = numpy.pi
        if errortheta > numpy.pi:
            errortheta = numpy.pi

        # The peak from "moments" is just the value of the maximum pixel
        # times a correction, fudge_max_pix, for the fact that the
        # centre of the Gaussian is not at the centre of the pixel.
        # This correction is performed in fitting.py. The maximum pixel
        # method introduces a peak dependent error corresponding to the last
        # term in the expression below for errorpeaksq.
        # To this, we add, in quadrature, the errors corresponding
        # to the first and last term of the rhs of equation 37 of the
        # NVSS paper. The middle term in that equation 37 is heuristically
        # replaced by noise**2 since the threshold should not affect
        # the error from the (corrected) maximum pixel method,
        # while it is part of the expression for rho_sq above.
        # errorpeaksq = ((frac_flux_cal_error * peak) ** 2 +
        #                clean_bias_error ** 2 + noise ** 2 +
        #                utils.maximum_pixel_method_variance(
        #                    beam[0], beam[1], beam[2]) * peak ** 2)
        errorpeaksq = ((frac_flux_cal_error * peak) ** 2 +
                       clean_bias_error ** 2 + noise ** 2 +
                       max_pix_variance_factor * peak ** 2)
        errorpeak = numpy.sqrt(errorpeaksq)

        help1 = (errorsmaj / smaj) ** 2
        help2 = (errorsmin / smin) ** 2
        help3 = theta_B * theta_b / (4. * smaj * smin)
        errorflux = flux * numpy.sqrt(
            errorpeaksq / peak ** 2 + help3 * (help1 + help2))

        self['peak'].error = errorpeak
        self['flux'].error = errorflux
        self['xbar'].error = errorx
        self['ybar'].error = errory
        self['semimajor'].error = errorsmaj
        self['semiminor'].error = errorsmin
        self['theta'].error = errortheta

        return self

    def deconvolve_from_clean_beam(self, beam):
        """Deconvolve with the clean beam"""

        # If the fitted axes are larger than the clean beam
        # (=restoring beam) axes, the axes and position angle
        # can be deconvolved from it.
        fmaj = 2. * self['semimajor'].value
        fmajerror = 2. * self['semimajor'].error
        fmin = 2. * self['semiminor'].value
        fminerror = 2. * self['semiminor'].error
        fpa = numpy.degrees(self['theta'].value)
        fpaerror = numpy.degrees(self['theta'].error)
        cmaj = 2. * beam[0]
        cmin = 2. * beam[1]
        cpa = numpy.degrees(beam[2])

        rmaj, rmin, rpa, ierr = deconv(fmaj, fmin, fpa, cmaj, cmin, cpa)
        # This parameter gives the number of components that could not be
        # deconvolved, IERR from deconf.f.
        self.deconv_imposs = ierr
        # Now, figure out the error bars.
        if rmaj > 0:
            # In this case the deconvolved position angle is defined.
            # For convenience we reset rpa to the interval [-90, 90].
            if rpa > 90:
                rpa = -numpy.mod(-rpa, 180.)
            self['theta_deconv'].value = rpa

            # In the general case, where the restoring beam is elliptic,
            # calculating the error bars of the deconvolved position angle
            # is more complicated than in the NVSS case, where a circular
            # restoring beam was used.
            # In the NVSS case the error bars of the deconvolved angle are
            # equal to the fitted angle.
            rmaj1, rmin1, rpa1, ierr1 = deconv(
                fmaj, fmin, fpa + fpaerror, cmaj, cmin, cpa)
            if ierr1 < 2:
                if rpa1 > 90:
                    rpa1 = -numpy.mod(-rpa1, 180.)
                rpaerror1 = numpy.abs(rpa1 - rpa)
                # An angle error can never be more than 90 degrees.
                if rpaerror1 > 90.:
                    rpaerror1 = numpy.mod(-rpaerror1, 180.)
            else:
                rpaerror1 = numpy.nan
            rmaj2, rmin2, rpa2, ierr2 = deconv(
                fmaj, fmin, fpa - fpaerror, cmaj, cmin, cpa)
            if ierr2 < 2:
                if rpa2 > 90:
                    rpa2 = -numpy.mod(-rpa2, 180.)
                rpaerror2 = numpy.abs(rpa2 - rpa)
                # An angle error can never be more than 90 degrees.
                if rpaerror2 > 90.:
                    rpaerror2 = numpy.mod(-rpaerror2, 180.)
            else:
                rpaerror2 = numpy.nan
            if numpy.isnan(rpaerror1) or numpy.isnan(rpaerror2):
                self['theta_deconv'].error = numpy.nansum(
                    [rpaerror1, rpaerror2])
            else:
                self['theta_deconv'].error = numpy.mean(
                    [rpaerror1, rpaerror2])
            self['semimaj_deconv'].value = rmaj / 2.
            rmaj3, rmin3, rpa3, ierr3 = deconv(
                fmaj + fmajerror, fmin, fpa, cmaj, cmin, cpa)
            # If rmaj>0, then rmaj3 should also be > 0,
            # if I am not mistaken, see the formulas at
            # the end of ch.2 of Spreeuw's Ph.D. thesis.
            if fmaj - fmajerror > fmin:
                rmaj4, rmin4, rpa4, ierr4 = deconv(
                    fmaj - fmajerror, fmin, fpa, cmaj, cmin, cpa)
                if rmaj4 > 0:
                    self['semimaj_deconv'].error = numpy.mean(
                        [numpy.abs(rmaj3 - rmaj), numpy.abs(rmaj - rmaj4)])
                else:
                    self['semimaj_deconv'].error = numpy.abs(rmaj3 - rmaj)
            else:
                rmin4, rmaj4, rpa4, ierr4 = deconv(
                    fmin, fmaj - fmajerror, fpa, cmaj, cmin, cpa)
                if rmaj4 > 0:
                    self['semimaj_deconv'].error = numpy.mean(
                        [numpy.abs(rmaj3 - rmaj), numpy.abs(rmaj - rmaj4)])
                else:
                    self['semimaj_deconv'].error = numpy.abs(rmaj3 - rmaj)
            if rmin > 0:
                self['semimin_deconv'].value = rmin / 2.
                if fmin + fminerror < fmaj:
                    rmaj5, rmin5, rpa5, ierr5 = deconv(
                        fmaj, fmin + fminerror, fpa, cmaj, cmin, cpa)
                else:
                    rmin5, rmaj5, rpa5, ierr5 = deconv(
                        fmin + fminerror, fmaj, fpa, cmaj, cmin, cpa)
                # If rmin > 0, then rmin5 should also be > 0,
                # if I am not mistaken, see the formulas at
                # the end of ch.2 of Spreeuw's Ph.D. thesis.
                rmaj6, rmin6, rpa6, ierr6 = deconv(
                    fmaj, fmin - fminerror, fpa, cmaj, cmin, cpa)
                if rmin6 > 0:
                    self['semimin_deconv'].error = numpy.mean(
                        [numpy.abs(rmin6 - rmin), numpy.abs(rmin5 - rmin)])
                else:
                    self['semimin_deconv'].error = numpy.abs(rmin5 - rmin)
            else:
                self['semimin_deconv'] = Uncertain(
                    numpy.nan, numpy.nan)
        else:
            self['semimaj_deconv'] = Uncertain(numpy.nan, numpy.nan)
            self['semimin_deconv'] = Uncertain(numpy.nan, numpy.nan)
            self['theta_deconv'] = Uncertain(numpy.nan, numpy.nan)

        return self


def source_profile_and_errors(data, threshold, noise,
                              beam, fudge_max_pix_factor, max_pix_variance_factor,
                              beamsize, correlation_lengths, fixed=None):
    """Return a number of measurable properties with errorbars

    Given an island of pixels it will return a number of measurable
    properties including errorbars.  It will also compute residuals
    from Gauss fitting and export these to a residual map.

    In addition to handling the initial parameter estimation, and any fits
    which fail to converge, this function runs the goodness-of-fit
    calculations -
    see :func:`sourcefinder.fitting.goodness_of_fit` for details.

    Args:

        data (numpy.ndarray): array of pixel values, can be a masked
            array, which is necessary for proper Gauss fitting,
            because the pixels below the threshold in the corners and
            along the edges should not be included in the fitting
            process

        threshold (float): Threshold used for selecting pixels for the
            source (ie, building an island)

        noise (float): Noise level in data

        beam (tuple): beam parameters (semimaj,semimin,theta)

        fudge_max_pix_factor(float): Correct for the underestimation of the peak
                                     by taking the maximum pixel value.

        max_pix_variance_factor(float): Take account of additional variance
                                        induced by the maximum pixel method,
                                        on top of the background noise.

        beamsize(float): The FWHM size of the clean beam

        correlation_lengths(tuple): Tuple of two floats describing the distance
                                    along the semimajor and semiminor axes of
                                    the clean beam beyond which noise is assumed
                                    uncorrelated. Some background: Aperture
                                    synthesis imaging yields noise that is
                                    partially correlated over the entire image.
                                    This has a considerable effect on error
                                    estimates. We approximate this by
                                    considering all noise within the
                                    correlation length completely correlated
                                     and beyond that completely uncorrelated.

    Kwargs:

        fixed (dict): Parameters (and their values) to hold fixed while fitting.
            Passed on to fitting.fitgaussian().

    Returns:
        tuple: a populated ParamSet, an islands map and a residuals map.
            Note that both the islands and residuals maps are regular ndarrays,
            where masked (unfitted) regions have been filled with 0-values.

    """

    if fixed is None:
        fixed = {}
    param = ParamSet()

    if threshold is None:
        moments_threshold = 0
    else:
        moments_threshold = threshold

    # We can always find the maximum pixel value and derive the barycenter
    # position. The other three Gaussian parameters we can copy from the clean
    # beam.
    # Are we fitting a -ve or +ve Gaussian?
    if data.mean() >= 0:
        # The peak is always underestimated when you take the highest pixel.
        peak = data.max() * fudge_max_pix_factor
    else:
        peak = data.min()
    total = data.sum()
    x, y = numpy.indices(data.shape)
    xbar = float((x * data).sum() / total)
    ybar = float((y * data).sum() / total)

    param.update({
        "peak": peak,
        "flux": peak,
        "xbar": xbar,
        "ybar": ybar,
        "semimajor": beam[0],
        "semiminor": beam[1],
        "theta": beam[2]
    })
    # data_as_ones is constructed to help determine if the island has enough
    # width for Gauss fitting.
    try:
        if data.mask.shape == data.data.shape:
            data_as_ones = numpy.where(~data.mask == 1, 1, 0)
        else:
            data_as_ones = numpy.where(data.data > moments_threshold, 1, 0)
    except AttributeError:
        data_as_ones = numpy.where(data > moments_threshold, 1, 0)
    max_along_x = numpy.sum(data_as_ones, axis=0).max()
    max_along_y = numpy.sum(data_as_ones, axis=1).max()
    minimum_width = min(max_along_x, max_along_y)

    if minimum_width > 2:
        # If the island or subisland has thickness of more than 2 in both
        # dimensions we can properly compute moments and fit a Gaussian with
        # six free parameters.
        try:
            # We can compute moments and possibly try fitting if the island has a
            # thickness of more than 2 in both dimensions.
            param.update(fitting.moments(data, fudge_max_pix_factor, beamsize,
                                         moments_threshold))
            param.moments = True
        except ValueError:
            logger.warning('Moments computations failed, use defaults.')
        try:
            param.compute_bounds(data.shape)
            gaussian_soln = fitting.fitgaussian(data, param, fixed=fixed,
                                                bounds=param.bounds)
            param.update(gaussian_soln)
            param.gaussian = True
            logger.debug('Gaussian fitting was successful.')
        except ValueError:
            logger.warning('Gaussian fitting failed.')
    else:
        logger.debug("Unable to estimate gaussian parameters from moments."
                     " Proceeding with defaults %s""",
                     str(param))

    if fixed and not param.gaussian:
        # moments can't handle fixed params
        raise ValueError("fit failed with given fixed parameters")

    param["flux"] = (numpy.pi * param["peak"] * param["semimajor"] *
                     param["semiminor"] / beamsize)
    param.calculate_errors(noise, max_pix_variance_factor, correlation_lengths, threshold)
    param.deconvolve_from_clean_beam(beam)

    # Calculate residuals
    # NB this works even if Gaussian fitting fails, we generate the model from
    # the moments-fit parameters.
    gauss_arg = (param["peak"].value,
                 param["xbar"].value,
                 param["ybar"].value,
                 param["semimajor"].value,
                 param["semiminor"].value,
                 param["theta"].value)

    try:
        gauss_island_masked = \
            numpy.ma.array(gaussian(*gauss_arg)(*numpy.indices(data.shape)),
                           mask=data.mask)
        gauss_resid_masked = data - gauss_island_masked
        gauss_island_filled = gauss_island_masked.filled(fill_value=0.)
        gauss_resid_filled = gauss_resid_masked.filled(fill_value=0.)
    except AttributeError:
        gauss_island_masked = gaussian(*gauss_arg)(*numpy.indices(data.shape))
        gauss_resid_masked = data - gauss_island_masked
        gauss_island_filled = gauss_island_masked
        gauss_resid_filled = gauss_resid_masked

    param.chisq, param.reduced_chisq = fitting.goodness_of_fit(
        gauss_resid_masked, noise, correlation_lengths)

    return param, gauss_island_filled, gauss_resid_filled


class Detection(object):
    """The result of a measurement at a given position in a given image."""

    def __init__(self, paramset, imagedata, chunk=None, eps_ra=0, eps_dec=0):

        self.eps_ra = eps_ra
        self.eps_dec = eps_dec

        self.imagedata = imagedata
        # self.wcs = imagedata.wcs
        self.chunk = chunk

        self.peak = paramset['peak']
        self.flux = paramset['flux']
        self.x = paramset['xbar']
        self.y = paramset['ybar']
        self.smaj = paramset['semimajor']
        self.smin = paramset['semiminor']
        self.theta = paramset['theta']
        # This parameter gives the number of components that could not
        # be deconvolved, IERR from deconf.f.
        self.dc_imposs = paramset.deconv_imposs
        self.smaj_dc = paramset['semimaj_deconv']
        self.smin_dc = paramset['semimin_deconv']
        self.theta_dc = paramset['theta_deconv']
        self.error_radius = None
        self.gaussian = paramset.gaussian
        self.chisq = paramset.chisq
        self.reduced_chisq = paramset.reduced_chisq

        self.sig = paramset.sig

        try:
            self._physical_coordinates()
        except RuntimeError:
            logger.warning("Physical coordinates failed at %f, %f" % (
                self.x, self.y))
            raise

    def __getstate__(self):
        return {
            'imagedata': self.imagedata,
            'chunk': (self.chunk[0].start, self.chunk[0].stop,
                      self.chunk[1].start, self.chunk[1].stop),
            'peak': self.peak,
            'flux': self.flux,
            'x': self.x,
            'y': self.y,
            'smaj': self.smaj,
            'smin': self.smin,
            'theta': self.theta,
            'sig': self.sig,
            'error_radius': self.error_radius,
            'gaussian': self.gaussian,
        }

    def __setstate__(self, attrdict):
        self.imagedata = attrdict['imagedata']
        self.chunk = (slice(attrdict['chunk'][0], attrdict['chunk'][1]),
                      slice(attrdict['chunk'][2], attrdict['chunk'][3]))
        self.peak = attrdict['peak']
        self.flux = attrdict['flux']
        self.x = attrdict['x']
        self.y = attrdict['y']
        self.smaj = attrdict['smaj']
        self.smin = attrdict['smin']
        self.theta = attrdict['theta']
        self.sig = attrdict['sig']
        self.error_radius = attrdict['error_radius']
        self.gaussian = attrdict['gaussian']

        try:
            self._physical_coordinates()
        except RuntimeError as e:
            logger.warning("Physical coordinates failed at %f, %f" % (
                self.x, self.y))
            raise

    def __getattr__(self, attrname):
        # Backwards compatibility for "errquantity" attributes
        if attrname[:3] == "err":
            return self.__getattribute__(attrname[3:]).error
        else:
            raise AttributeError(attrname)

    def __str__(self):
        return "(%.2f, %.2f) +/- (%.2f, %.2f): %g +/- %g" % (
            self.ra.value, self.dec.value, self.ra.error * 3600,
            self.dec.error * 3600,
            self.peak.value, self.peak.error)

    def __repr__(self):
        return str(self)

    def _physical_coordinates(self):
        """Convert the pixel parameters for this object into something
        physical."""

        # First, the RA & dec.
        self.ra, self.dec = [Uncertain(x) for x in self.imagedata.wcs.p2s(
            [self.x.value, self.y.value])]
        if numpy.isnan(self.dec.value) or abs(self.dec) > 90.0:
            raise ValueError("object falls outside the sky")

        # First, determine local north.
        help1 = numpy.cos(numpy.radians(self.ra.value))
        help2 = numpy.sin(numpy.radians(self.ra.value))
        help3 = numpy.cos(numpy.radians(self.dec.value))
        help4 = numpy.sin(numpy.radians(self.dec.value))
        center_position = numpy.array([help3 * help1, help3 * help2, help4])

        # The length of this vector is chosen such that it touches
        # the tangent plane at center position.
        # The cross product of the local north vector and the local east
        # vector will always be aligned with the center_position vector.
        if center_position[2] != 0:
            local_north_position = numpy.array(
                [0., 0., 1. / center_position[2]])
        else:
            # If we are right on the equator (ie dec=0) the division above
            # will blow up: as a workaround, we use something Really Big
            # instead.
            local_north_position = numpy.array([0., 0., 99e99])
        # Next, determine the orientation of the y-axis wrt local north
        # by incrementing y by a small amount and converting that
        # to celestial coordinates. That small increment is conveniently
        # chosen to be an increment of 1 pixel.

        endy_ra, endy_dec = self.imagedata.wcs.p2s(
            [self.x.value, self.y.value + 1.])
        help5 = numpy.cos(numpy.radians(endy_ra))
        help6 = numpy.sin(numpy.radians(endy_ra))
        help7 = numpy.cos(numpy.radians(endy_dec))
        help8 = numpy.sin(numpy.radians(endy_dec))
        endy_position = numpy.array([help7 * help5, help7 * help6, help8])

        # Extend the length of endy_position to make it touch the plane
        # tangent at center_position.
        endy_position /= numpy.dot(center_position, endy_position)

        diff1 = endy_position - center_position
        diff2 = local_north_position - center_position

        cross_prod = numpy.cross(diff2, diff1)

        length_cross_sq = numpy.dot(cross_prod, cross_prod)

        normalization = numpy.dot(diff1, diff1) * numpy.dot(diff2, diff2)

        # The length of the cross product equals the product of the lengths of
        # the vectors times the sine of their angle.
        # This is the angle between the y-axis and local north,
        # measured eastwards.
        # yoffset_angle = numpy.degrees(
        #    numpy.arcsin(numpy.sqrt(length_cross_sq/normalization)))
        # The formula above is commented out because the angle computed
        # in this way will always be 0<=yoffset_angle<=90.
        # We'll use the dotproduct instead.
        yoffs_rad = (numpy.arccos(numpy.dot(diff1, diff2) /
                                  numpy.sqrt(normalization)))

        # The multiplication with -sign_cor makes sure that the angle
        # is measured eastwards (increasing RA), not westwards.
        sign_cor = (numpy.dot(cross_prod, center_position) /
                    numpy.sqrt(length_cross_sq))
        yoffs_rad *= -sign_cor
        yoffset_angle = numpy.degrees(yoffs_rad)

        # Now that we have the BPA, we can also compute the position errors
        # properly, by projecting the errors in pixel coordinates (x and y)
        # on local north and local east.
        errorx_proj = numpy.sqrt(
            (self.x.error * numpy.cos(yoffs_rad)) ** 2 +
            (self.y.error * numpy.sin(yoffs_rad)) ** 2)
        errory_proj = numpy.sqrt(
            (self.x.error * numpy.sin(yoffs_rad)) ** 2 +
            (self.y.error * numpy.cos(yoffs_rad)) ** 2)

        # Now we have to sort out which combination of errorx_proj and
        # errory_proj gives the largest errors in RA and Dec.
        try:
            end_ra1, end_dec1 = self.imagedata.wcs.p2s(
                [self.x.value + errorx_proj, self.y.value])
            end_ra2, end_dec2 = self.imagedata.wcs.p2s(
                [self.x.value, self.y.value + errory_proj])
            # Here we include the position calibration errors
            self.ra.error = self.eps_ra + max(
                numpy.fabs(self.ra.value - end_ra1),
                numpy.fabs(self.ra.value - end_ra2))
            self.dec.error = self.eps_dec + max(
                numpy.fabs(self.dec.value - end_dec1),
                numpy.fabs(self.dec.value - end_dec2))
        except RuntimeError:
            # We get a runtime error from wcs.p2s if the errors place the
            # limits outside the image.
            # In which case we set the RA / DEC uncertainties to infinity
            self.ra.error = float('inf')
            self.dec.error = float('inf')

        # Estimate an absolute angular error on our central position.
        self.error_radius = utils.get_error_radius(
            self.imagedata.wcs, self.x.value, self.x.error, self.y.value,
            self.y.error
        )

        # Now we can compute the BPA, east from local north.
        # That these angles can simply be added is not completely trivial.
        # First, the Gaussian in gaussian.py must be such that theta is
        # measured from the positive y-axis in the direction of negative x.
        # Secondly, x and y are defined such that the direction
        # positive y-->negative x-->negative y-->positive x is the same
        # direction (counterclockwise) as (local) north-->east-->south-->west.
        # If these two conditions are matched, the formula below is valid.
        # Of course, the formula is also valid if theta is measured
        # from the positive y-axis towards positive x
        # and both of these directions are equal (clockwise).
        self.theta_celes = Uncertain(
            (numpy.degrees(self.theta.value) + yoffset_angle) % 180,
            numpy.degrees(self.theta.error))
        if not numpy.isnan(self.theta_dc.value):
            self.theta_dc_celes = Uncertain(
                (self.theta_dc.value + yoffset_angle) % 180,
                numpy.degrees(self.theta_dc.error))
        else:
            self.theta_dc_celes = Uncertain(numpy.nan, numpy.nan)

        # Next, the axes.
        # Note that the signs of numpy.sin and numpy.cos in the
        # four expressions below are arbitrary.
        self.end_smaj_x = (self.x.value - numpy.sin(self.theta.value) *
                           self.smaj.value)
        self.start_smaj_x = (self.x.value + numpy.sin(self.theta.value) *
                             self.smaj.value)
        self.end_smaj_y = (self.y.value + numpy.cos(self.theta.value) *
                           self.smaj.value)
        self.start_smaj_y = (self.y.value - numpy.cos(self.theta.value) *
                             self.smaj.value)
        self.end_smin_x = (self.x.value + numpy.cos(self.theta.value) *
                           self.smin.value)
        self.start_smin_x = (self.x.value - numpy.cos(self.theta.value) *
                             self.smin.value)
        self.end_smin_y = (self.y.value + numpy.sin(self.theta.value) *
                           self.smin.value)
        self.start_smin_y = (self.y.value - numpy.sin(self.theta.value) *
                             self.smin.value)

        def pixel_to_spatial(x, y):
            try:
                return self.imagedata.wcs.p2s([x, y])
            except RuntimeError:
                logger.debug("pixel_to_spatial failed at %f, %f" % (x, y))
                return numpy.nan, numpy.nan

        end_smaj_ra, end_smaj_dec = pixel_to_spatial(self.end_smaj_x,
                                                     self.end_smaj_y)
        end_smin_ra, end_smin_dec = pixel_to_spatial(self.end_smin_x,
                                                     self.end_smin_y)

        smaj_asec = coordinates.angsep(self.ra.value, self.dec.value,
                                       end_smaj_ra, end_smaj_dec)
        scaling_smaj = smaj_asec / self.smaj.value
        errsmaj_asec = scaling_smaj * self.smaj.error
        self.smaj_asec = Uncertain(smaj_asec, errsmaj_asec)

        smin_asec = coordinates.angsep(self.ra.value, self.dec.value,
                                       end_smin_ra, end_smin_dec)
        scaling_smin = smin_asec / self.smin.value
        errsmin_asec = scaling_smin * self.smin.error
        self.smin_asec = Uncertain(smin_asec, errsmin_asec)

    def distance_from(self, x, y):
        """Distance from center"""
        return ((self.x - x) ** 2 + (self.y - y) ** 2) ** 0.5

    def serialize(self, ew_sys_err, ns_sys_err):
        """
        Return source properties suitable for database storage.

        We manually add ew_sys_err, ns_sys_err

        returns: a list of tuples containing all relevant fields
        """
        return [
            self.ra.value,
            self.dec.value,
            self.ra.error,
            self.dec.error,
            self.peak.value,
            self.peak.error,
            self.flux.value,
            self.flux.error,
            self.sig,
            self.smaj_asec.value,
            self.smin_asec.value,
            self.theta_celes.value,
            ew_sys_err,
            ns_sys_err,
            self.error_radius,
            self.gaussian,
            self.chisq,
            self.reduced_chisq
        ]


@guvectorize([(float64[:], float64[:], float32[:], float32[:], float32[:],
               float32[:])], '(n), (n), (n), (l), (m) -> (m)', nopython=True)
def first_part_of_celestial_coordinates(ra_dec, endy_ra_dec,
                                        xbar_ybar_error,
                                        xbar_ybar_smaj_smin_theta,
                                        dummy, return_values):
    """
    Similar to extract.Detection._physical_coordinates, but vectorized and
    mainly the first part, until we need another call to wcs.all_pix2world,
    based on the output from this part.
    What we have learned from moments_enhanced is that measuring the islands
    can be done very rapidly, by vectorizing the measurements rather than
    parallellizing, although one can do both by using the guvectorize
    decorator from Numba. Contrary to vectorizing moments, vectorizing
    the transformation to celestial coordinates is harder, since it involves
    wcs.wcs_pix2world from Astropy, which Numba cannot compile. The way out
    of this is to use wcs.all_pix2world from Astropy, which is a fast
    vectorized function, but this has to be done outside the routine compiled
    by Numba, i.e. this routine. We will be needing a number of calls to
    wcs.all_pix2world to traverse all the steps from
    extract.Detection._physical_coordinates.

    Args:
        ra_dec (numpy.ndarray): array of floats of length 2, containing the
            right ascension (degrees) and the declination
            (degrees) corresponding to [xbar, ybar] of the source.

        endy_ra_dec (numpy.ndarray): array of floats of length 2, containing
            the right ascension (degrees) and the declination
            (degrees) corresponding to [xbar, ybar + 1] of the source.

        xbar_ybar_error (numpy.ndarray): array of floats of length 2,
            with the errors on the barycentric positions, in both dimensions.

        xbar_ybar_smaj_smin_theta (numpy.ndarray): array of floats of length 5,
            with the barycentric positions, the semi-major and semi-minor axes
            and the position angles.

    Returns:
        None (because of the guvectorize decorator), but return_values is
        filled with xerror_proj, yerror_proj, yoffset_angle, end_smaj_x,
        start_smaj_x, end_smaj_y,  start_smaj_y, end_smin_x, start_smin_x,
        end_smin_y and start_smin_y.
    """

    ra, dec = ra_dec
    if numpy.isnan(dec) or abs(dec) > 90.0:
        raise ValueError("object falls outside the sky")

    # First, determine local north.
    help1 = numpy.cos(numpy.radians(ra))
    help2 = numpy.sin(numpy.radians(ra))
    help3 = numpy.cos(numpy.radians(dec))
    help4 = numpy.sin(numpy.radians(dec))
    center_position = numpy.array([help3 * help1, help3 * help2, help4])

    # The length of this vector is chosen such that it touches
    # the tangent plane at center position.
    # The cross product of the local north vector and the local east
    # vector will always be aligned with the center_position vector.
    if center_position[2] != 0:
        local_north_position = numpy.array(
            [0., 0., 1. / center_position[2]])
    else:
        # If we are right on the equator (ie dec=0) the division above
        # will blow up: as a workaround, we use something Really Big
        # instead.
        local_north_position = numpy.array([0., 0., 99e99])

    endy_ra, endy_dec = endy_ra_dec
    help5 = numpy.cos(numpy.radians(endy_ra))
    help6 = numpy.sin(numpy.radians(endy_ra))
    help7 = numpy.cos(numpy.radians(endy_dec))
    help8 = numpy.sin(numpy.radians(endy_dec))
    endy_position = numpy.array([help7 * help5, help7 * help6, help8])

    # Extend the length of endy_position to make it touch the plane
    # tangent at center_position.
    endy_position /= numpy.dot(center_position, endy_position)

    diff1 = endy_position - center_position
    diff2 = local_north_position - center_position

    cross_prod = numpy.cross(diff2, diff1)

    length_cross_sq = numpy.dot(cross_prod, cross_prod)

    normalization = numpy.dot(diff1, diff1) * numpy.dot(diff2, diff2)

    # The length of the cross product equals the product of the lengths of
    # the vectors times the sine of their angle.
    # This is the angle between the y-axis and local north,
    # measured eastwards.
    # yoffset_angle = numpy.degrees(
    #    numpy.arcsin(numpy.sqrt(length_cross_sq/normalization)))
    # The formula above is commented out because the angle computed
    # in this way will always be 0<=yoffset_angle<=90.
    # We'll use the dotproduct instead.
    yoffs_rad = (numpy.arccos(numpy.dot(diff1, diff2) /
                              numpy.sqrt(normalization)))

    # The multiplication with -sign_cor makes sure that the angle
    # is measured eastwards (increasing RA), not westwards.
    sign_cor = (numpy.dot(cross_prod, center_position) /
                numpy.sqrt(length_cross_sq))
    yoffs_rad *= -sign_cor
    yoffset_angle = numpy.degrees(yoffs_rad)

    # Now that we have the BPA, we can also compute the position errors
    # properly, by projecting the errors in pixel coordinates (x and y)
    # on local north and local east.
    xbar_error, ybar_error = xbar_ybar_error
    errorx_proj = numpy.sqrt(
        (xbar_error * numpy.cos(yoffs_rad)) ** 2 +
        (ybar_error * numpy.sin(yoffs_rad)) ** 2)
    errory_proj = numpy.sqrt(
        (xbar_error * numpy.sin(yoffs_rad)) ** 2 +
        (ybar_error * numpy.cos(yoffs_rad)) ** 2)

    # Next, the axes.
    # Note that the signs of numpy.sin and numpy.cos in the
    # four expressions below are arbitrary.
    xbar, ybar, smaj, smin, theta = xbar_ybar_smaj_smin_theta
    end_smaj_x = xbar - numpy.sin(theta) * smaj
    start_smaj_x = xbar + numpy.sin(theta) * smaj
    end_smaj_y = ybar + numpy.cos(theta) * smaj
    start_smaj_y = ybar - numpy.cos(theta) * smaj
    end_smin_x = xbar + numpy.cos(theta) * smin
    start_smin_x = xbar - numpy.cos(theta) * smin
    end_smin_y = ybar + numpy.sin(theta) * smin
    start_smin_y = ybar - numpy.sin(theta) * smin

    return_values[:] = numpy.array([errorx_proj, errory_proj, yoffset_angle,
                                    end_smaj_x, start_smaj_x, end_smaj_y, start_smaj_y,
                                    end_smin_x, start_smin_x, end_smin_y, start_smin_y])


@guvectorize([(float32[:, :], int32[:], int32[:, :], int32[:], int32[:],
               int32[:], float32[:], int32[:], int32[:], int32[:])],
             '(n, m), (l), (n, m), (), (), (k) -> (k), (k), (k), ()')
def insert_island_data(some_image, inds, labelled_data, label,
                       npix, dummy, island, xpos, ypos, min_width):
    """
    We want to copy the relevant island data into input arrays for
    fitting.moments_enhanced.

    :param some_image: The 2D ndarray with all the pixel values, typically
                       self.data_bgsubbed.data.

    :param inds: A ndarray of four indices indicating the slice encompassing
                 an island. Such a slice would typically be a pick from a
                 list of slices from a call to scipy.ndimage.find_objects.
                 Since we are attempting vectorized processing here, the
                 slice should have been replaced by its four coordinates
                 through a call to slices_to_indices.

    :param labelled_data: A ndarray with the same shape as some_image, with
                          labelled islands with integer values and zeroes
                          for all background pixels.

    :param label: The label (integer value) corresponding to the slice
                  encompassing the island. Or actually it should be the
                  other way round, since there can be multiple islands
                  within one rectangular slice.

    :param npix: Number of pixels comprising the island with label=label.

    :param dummy: Artefact of the implementation of guvectorize:
                  Empty 1D ndarray with the same length as island, xpos and
                  ypos, i.e. maxpix. Defined as int32 array, but could be
                  any other type. It is needed because of a missing feature
                  in guvectorize: there is no other way to tell guvectorize
                  what the shape of the output array will be. Therefore, we
                  define an otherwise redundant input array with the same
                  shape as the desired output array.

    :param island:  1D ndarray of float32 numbers of pixel values of the
                    island with label = label. Length = maxpix, the number
                    of pixels in the largest possible island, indicating
                    that there will be a number of unassigned values for
                    the highest indices.

    :param xpos: 1D ndarray of integers indicating the row indices of the
                 pixels of the island with label = label, relative to the
                 position of pixel [0, 0] of the rectangular slice
                 encompassing the island. Must have same order as pixel
                 values in island. Length = maxpix, the number of pixels in
                 the largest possible island, indicating that there will be
                 a number of unassigned values for the highest indices.

    :param ypos: 1D ndarray of integers indicating the column indices of the
                 pixels of the island with label = label, relative to the
                 position of pixel [0, 0]  of the rectangular slice
                 encompassing the island. Must have same order as
                 pixel values in island. Length = maxpix, the number of
                 pixels in the largest possible island, indicating that
                 there will be a number of unassigned values for the highest
                 indices.

    :param min_width: int32 indicating the minimum width of the island, in order
                 to assess whether the island has sufficient width to determine
                 Gaussian parameters. Calculated from the maximum widths of the
                 island over x and y and subsequently taking the minimum of
                 those two maximum widths.

    :return: No return values, because of the use of the guvectorize
             decorator: 'guvectorize() functions don’t return their
             result value: they take it as an array argument,
             which must be filled in by the function'. In this case
             island, xpos, ypos and min_width will be filled with values.
    """

    labelled_data_chunk = labelled_data[inds[0]:inds[1], inds[2]:inds[3]]
    segmented_island = numpy.where(labelled_data_chunk == label[0], 1, 0)

    max_along_x = numpy.sum(segmented_island, axis=0).max()
    max_along_y = numpy.sum(segmented_island, axis=1).max()
    min_width[0] = min(max_along_x, max_along_y)

    # pos = "positions", i.e. the row and column indices of the island pixels.
    pos = segmented_island.nonzero()
    image_chunk = some_image[inds[0]:inds[1], inds[2]:inds[3]]

    for i in range(npix[0]):
        index = pos[0][i], pos[1][i]
        island[i] = image_chunk[index]
        xpos[i], ypos[i] = index


def source_measurements_pixels_and_celestial_vectorised(num_islands, npixs,
                                                        maxposs, maxis,
                                                        data_bgsubbeddata,
                                                        rmsdata,
                                                        analysisthresholddata,
                                                        indices,
                                                        labelled_data, labels, wcs,
                                                        fudge_max_pix_factor,
                                                        max_pix_variance_factor,
                                                        beam, beamsize,
                                                        correlation_lengths,
                                                        eps_ra, eps_dec):
    """
    From islands of pixels above the analysis threshold with peaks above the
    detection threshold source parameters are extracted, including error bars
    These quantities are transformed to celestial coordinates in a vectorized
    way, making it suitable for tens of thousands of sources per image.
    Per island one can extract a maximum of 10 quantities: peak spectral
    brightness, flux density, i.e. the spectral brightness integrated over
    the source, position in the sky along both axes and the source shape
    parameters: semi-major and semi-minor axes and the position angle of the
    major axis, measured east from local north. The latter three quantities can
    be deconvolved from the clean beam if the source is resolved. That gives a
    total af 20 numbers, 10 measurements and their error bars. 8 of these
    have to be transformed to celestial coordinates, which adds another 16
    numbers.
    This function is a duplicate of code elsewhere, in particular of
    'fitting.moments' and 'extract.Detection._physical_coordinates' but with
    a data layout suitable for vectorisation.

    :param eps_ra:
    :param num_islands: integer equal to the number of islands detected.

    :param npixs: 1D int32 ndarray of length num_islands representing the number
                  of pixels comprising every island.

    :param maxposs: 2D int32 ndarray of shape (num_islands, 2) with the peak
                    positions per island

    :param maxis: 1D float32 ndarray of length num_islands with the peak pixel
                  values corresponding to the maxposs positions.

    :param data_bgsubbeddata: 2D ndarray of float32 or float64, representing the
                              image. These are the data of
                              ImageData.data_bgsubbed, i.e.
                              ImageData.data_bgsubbed.data, without the mask,
                              since we assume that any mask has already
                              been properly propagated in selecting the islands
                              pixels. The interpolated mean background has been
                              subtracted from the image data.

    :param rmsdata: 2D ndarray of float32 or float64, same shape as the image
                    (data_bgsubbeddata). rms is jargon in image processing, it
                    refers to "Root mean square", but is taken as the
                    standard deviation of the background noise, assuming
                    zero mean. The standard deviation of the background noise
                    is determined after kappa, sigma clipping subimages of
                    appropriate size and interpolating this across the image
                    while including masks. This is ImageData.rmsmap.data, i.e.
                    without the mask,since we assume that any mask has already
                    been properly propagated in selecting the islands pixels.

    :param analysisthresholddata: 2D ndarray of float32 or float64, same shape
                                  as data_bgsubbeddata. These are the data of
                                  the analysisthresholdmap, i.e.
                                  analysisthresholdmap.data, without the mask,
                                  since we assume that any mask has already
                                  been properly propagated in the process
                                  of selecting the island pixels.

    :param indices: 2D int32 ndarray of shape (num_islands, 4) representing
                    positions of the corners of the "chunks", i.e. slices
                    encompassing every island, relative to upper left corner
                    of the image, i.e. relative to array element [0,0] of
                    data_bgsubbeddata.

    :param labelled_data: 2D int32 ndarray of shape data_bgsubbeddata.shape,
                          representing all unmasked island pixels above the
                          analysis threshold, with labels (integers) > 0.
                          All background or masked pixels from
                          ImageData.data_bgsubbed should have value (label) = 0.

    :param labels: 1D int32 or int64 ndarray representing the labels in
                   labelled_data corresponding to islands in
                   data_bgsubbeddata that have peak values above the local
                   detection threshold. Thus, it is a subset of
                   numpy.unique(labelled_data) because the
                   'insignificant' islands have been filtered out, i.e.
                   the islands above the analysis threshold with peak
                   values below the (local) detection threshold. Also,
                   labels does not contain the zero label corresponding
                   to background or masked pixels in ImageData.data_bgsubbed.

    :param wcs:  A utility.coordinates.wcs instance, filled with appropriate
                 values from the image header, issued e.g. through
                 accessors.fitsimage.FitSImage. With numbers representing
                 the center of the image and the width of the pixels in
                 celestial coordinates plus the type of projection used,
                 we can transform from pixel coordinates to celestial
                 coordinates.

    :param fudge_max_pix_factor: float to correct for the underestimate of the
                 true peak when the maximum pixel method is used. It is a
                 statistical correction, so on average correct over a large
                 ensemble of unresolved sources, when a circular restoring beam
                 is appropriate.

    :param max_pix_variance_factor: float to take account of the additional
             uncertainty introduced by fudge_max_pix. Can probably be removed
             since its effect is negligible in all sensible cases.

    :param beam: tuple of 3 floats describing the restoring beam in terms of
                 its semi-major and semi-minor axes (in pixels) and the position
                 angle of the semi-major axis, east from local north, in
                 radians.

    :param beamsize: area (float) of the restoring beam, in pixels.

    :param correlation_lengths: tuple of 2 floats describing over which
             distance (in pixels) noise should be considered correlated, along
             both principal axes of the Gaussian profile of the restoring
             beam.

    :param eps_ra, eps_dec: floats (degrees) that reflect an extra positional
            uncertainty from calibration errors along right ascension and
            declination.

    :return:
             moments_of_sources: a 3D float32 ndarray, with a row index
             corresponding to each island, i.e. num_islands rows, a 0 or 1
             column index corresponding to values and their error bars,
             respectively and a third index [0...9] which represents the
             measurements of the sources, i.e. peak specific brightness,
             in Jy/beam and the integrated specific brightness a.k.a. flux
             density in Jy. All the other numbers in moments_of_sources are
             in pixel coordinates. These are position on the sky, along both
             axes, semi-major and semi-minor axis of the Gaussian profile of
             the source and the position angle of the major axis measured east
             from local north. These three Gaussian profile parameters can be
             deconvolved from the clean beam if the source is resolved. This
             adds up to ten quantities and corresponding error bars for each
             island.

             sky_barycenters: a (num_islands, 2) ndarray of floats: each row
                              has an entry for Right ascension (float64) and
                              Declination (float64).

             chunk_positions: a (num_islands, 2) ndarray of integers denoting
                              the indices that correspond to the upper left
                              corners of the islands. They are used in this
                              module and returned to the calling module.
                              However, their return is only useful if the user
                              requires a residual map.

             x_positions: a (num_islands, max_pixels) ndarray of integers
                          denoting the row indices of the island relative to
                          the upper left corner of the island. They are used in
                          this module and returned to the calling module.
                          However, their return is only useful if the user
                          requires a residual map.

             y_positions: a (num_islands, max_pixels) ndarray of integers
                          denoting the column indices of the island relative to
                          the upper left corner of the island. They are used in
                          this module and returned to the calling module.
                          However, their return is only useful if the user
                          requires a residual map.

             ra_errors, dec_errors: Both are 1D float64 ndarrays corresponding
                                    to the two columns in sky_barycenters.

             error_radii: a 1D float64 ndarray representing an absolute error
                          bar on the position of every source. This is an
                          important quantity to associate an observation of a
                          source with its previous observations.

             smaj_asec, errsmaj_asec, smin_asec, errsmin_asec: Four 1D float64
                          ndarrays representing the convolved semi-major and
                          semi-minor axes - i.e. not deconvolved from the
                          restoring beam - transformed from pixel coordinates
                          to celestial coordinates and their 1-sigma error bars.

             theta_celes_values and theta_celes_errors:  Both are 1D float32
                          ndarrays representing the position angle of the
                          semi-major axis of the convolved Gaussian profile,
                          i.e. not yet deconvolved from the restoring beam and
                          its estimated 1-sigma error bar.

             theta_dc_celes_values and theta_dc_celes_errors: Both are 1D
                          float32 ndarrays representing the position angle of
                          the semi-major axis of the deconvolved Gaussian
                          profile, i.e. deconvolved from the restoring beam and
                          its estimated 1-sigma error bar. The function we are
                          describing here is mimicking the Detection class as
                          closely as possible, including the
                          ._physical_coordinates function, that is why we are
                          including these two quantities. Oddly enough, these
                          are not propagated into the serialize function, which
                          aggregates quantities for database storage, nor are
                          the deconvolved semi-major and semi-minor axes
                          transformed to celestial coordinates. It would be
                          straightforward to calculate these and to include
                          them for database storage, together with
                          theta_dc_celes_values and theta_dc_celes_errors. A
                          reason not to pass them on for database storage may
                          be that most sources are unresolved, so the
                          deconvolved source shape parameters will be nans.

    """
    # This is the conditional route to the fastest algorithm for source
    # measurements, with no forced_beam and deblending options and no
    # Gauss fitting, so a coarser way of measuring sources.

    # Make 2D input data suitable for moments_enhanced with guvectorize
    # decorator, Each row will contain the pixel values of a flattened
    # island.
    # First, determine the correct dimensions for the input
    # The number of rows will be num_labels = number of islands
    # (sources).
    # The number of columns will be max_pixels = the maximum number
    # of pixels an island can have.
    # We will also be needing auxiliary arrays.
    # One will be an index array, with an extra dimension
    # (num_labels, max_pixels, 2).
    # This will be a 3D array of integers, indicating the relative
    # positions of the source pixel relative to a corner of the slice
    # enclosing the island.
    # Later we will be needing another 2D array of floats with a number
    # of quantities related to sky position.

    max_pixels = npixs.max()

    islands = numpy.empty((num_islands, max_pixels), dtype=numpy.float32)
    # In order to convert to celestial coordinates, at a later stage, we
    # need to keep a record of the positions of the upper left corners
    # of the chunks. moments_enhanced starts by calculating xbar and
    # ybar as if those upper left corners have indices [0, 0] in the
    # image. We can add the indices of the chunks as arguments of
    # moments_enhanced to correct for that.
    chunk_positions = numpy.empty((num_islands, 2), dtype=numpy.int32)
    # xpositions and ypositions are relative to the upper left corner of
    # the chunk.
    xpositions = numpy.empty((num_islands, max_pixels), dtype=numpy.int32)
    ypositions = numpy.empty((num_islands, max_pixels), dtype=numpy.int32)

    # It makes sense to calculate the minimum width of each island of pixels
    # upfront such that we are not trying to estimate any Gaussian parameters,
    # either through moments estimation or through fitting, when the island has
    # a width of less than 3 pixels along any axis. For these "thin" detections
    # we can still determine the peak spectral brightness and position, but the
    # other four Gaussian parameters, i.e. the flux density, the axes and the
    # position angle will have to be derived with the help of the clean beam
    # parameters, i.e. by assuming the source is unresolved.
    minimum_widths = numpy.empty(num_islands, dtype=numpy.int32)

    # Make sure we get a single precision array of floats, which
    # fitting.moments_enhanced expects.
    thresholds = analysisthresholddata[maxposs[:, 0],
                                       maxposs[:, 1]].astype(
                                       numpy.float32, copy=False)

    local_noise_levels = rmsdata[maxposs[:, 0],
                                 maxposs[:, 1]].astype(
                                 numpy.float32, copy=False)

    chunk_positions[:, 0] = indices[:, 0]
    chunk_positions[:, 1] = indices[:, 2]

    dummy = numpy.empty_like(xpositions)

    insert_island_data(data_bgsubbeddata.astype(
        dtype=numpy.float32, copy=False),
        indices, labelled_data, labels.astype(
            dtype=numpy.int32, copy=False), npixs,
        dummy, islands, xpositions, ypositions, minimum_widths)
    # The result will be put in an array 'moments_of_sources' containing
    # ten quantities and their uncertainties: peak flux density,
    # integrated flux, xbar, ybar, semi-major axis, semi-minor axis,
    # gaussian position angle and the deconvolved equivalents of the latter
    # three quantities.
    moments_of_sources = numpy.empty((num_islands, 2, 10),
                                     dtype=numpy.float32)
    dummy = numpy.empty_like(moments_of_sources)

    # This is a workaround for an unresolved issue:
    # https://github.com/numba/numba/issues/6690
    # The output shape can apparently not be set as fixed numbers.
    # So we will add a dummy array with shape corresponding
    # to the output array (moments_of_sources), as (useless) input
    # array. In this way Numba can infer the shape of the output array.
    fitting.moments_enhanced(islands, chunk_positions, xpositions,
                             ypositions, minimum_widths, npixs,
                             thresholds, local_noise_levels,
                             maxis, fudge_max_pix_factor,
                             max_pix_variance_factor,
                             numpy.array(beam),
                             beamsize,
                             numpy.array(correlation_lengths),
                             0, 0, dummy, moments_of_sources)

    barycentric_pixel_positions = moments_of_sources[:, 0, 2:4]
    # Convert the barycentric positions to celestial_coordinates.
    sky_barycenters = wcs.all_p2s(barycentric_pixel_positions)
    # We need to determine the orientation of the y-axis wrt local north
    # by incrementing y by a small amount and converting that
    # to celestial coordinates. That small increment is conveniently
    # chosen to be an increment of 1 pixel.
    endy_barycentric_positions = barycentric_pixel_positions.copy()
    endy_barycentric_positions[:, 1] += 1
    endy_sky_coordinates = wcs.all_p2s(endy_barycentric_positions)

    input_for_second_part = \
        numpy.empty((num_islands, 11), dtype=numpy.float32)
    # Unfortunately, the use of the guvectorize decorator again
    # requires a dummy input with the same shape as the output,
    # such that Numba can infer the shape of the output array.
    dummy = numpy.empty_like(input_for_second_part)

    first_part_of_celestial_coordinates(sky_barycenters,
                                        endy_sky_coordinates,
                                        moments_of_sources[:, 1, 2:4],
                                        moments_of_sources[:, 0, 2:7],
                                        dummy, input_for_second_part)

    # Derive an absolute angular error on position, as
    # utils.get_error_radius does, but less involved.
    # Simply derive the angle between the celestial positions
    # corresponding to [xbar, ybar] and [xbar + errorx,
    # ybar + errory], that should suffice.
    error_radii = numpy.empty(num_islands, dtype=numpy.float64)
    try:
        # Compute sky positions corresponding to [x_bar + x_error,
        #                                         y_bar + y_error]
        pix_offset = moments_of_sources[:, 0, 2:4] + \
                     moments_of_sources[:, 1, 2:4]
        sky_offset = wcs.all_p2s(pix_offset)
        coordinates.angsep_vectorized(sky_barycenters, sky_offset,
                                      error_radii)

    except RuntimeError:
        # Mimic error handling from utils.get_error_radius.
        # The downside is that all error radii will be infinite
        # also when only the calculation of the error radius for a
        # single source gives a RuntimeError.
        error_radii.fill(float('inf'))

    # Now we have to sort out which combination of errorx_proj and
    # errory_proj gives the largest errors in RA and Dec.
    try:
        # Derive end_ra1, end_dec1, end_ra2, end_dec2 as in
        # extract.Detection._physical_coordinates.
        # We need to add errorx_proj to xbar and zero to ybar.
        # So we need to extract the first column from
        # input_for_second_part and append a column with zeros
        # in order to perform the addition.
        helper1 = input_for_second_part[:, :1]
        errorx_proj_and_zeros = numpy.hstack((helper1,
                                              numpy.zeros((helper1.shape[0], 1),
                                                          dtype=helper1.dtype)))
        pix_x_plus_errorx_proj = (moments_of_sources[:, 0, 2:4] +
                                  errorx_proj_and_zeros)
        end_ra1_end_dec1 = wcs.all_p2s(pix_x_plus_errorx_proj)

        # Now we need to add errory_proj to ybar and zero to xbar.
        # So we need to extract the second column from
        # input_for_second_part and prepend a column with zeros
        # in order to perform the addition.
        helper2 = input_for_second_part[:, 1:2]
        zeros_and_errory_proj = numpy.hstack((
            numpy.zeros((helper2.shape[0], 1), dtype=helper2.dtype),
            helper2))
        pix_y_plus_errory_proj = moments_of_sources[:, 0, 2:4] + \
            zeros_and_errory_proj

        end_ra2_end_dec2 = wcs.all_p2s(pix_y_plus_errory_proj)

        # Here we include the position calibration errors
        ra_errors = eps_ra + numpy.maximum(
            numpy.fabs(sky_barycenters[:, :1] - end_ra1_end_dec1[:, :1]),
            numpy.fabs(sky_barycenters[:, :1] - end_ra2_end_dec2[:, :1]))

        dec_errors = eps_dec + numpy.maximum(
            numpy.fabs(sky_barycenters[:, 1:2] - end_ra1_end_dec1[:, 1:2]),
            numpy.fabs(sky_barycenters[:, 1:2] - end_ra2_end_dec2[:, 1:2]))
    except RuntimeError:
        # We get a runtime error from wcs.all_p2s if the errors place the
        # limits outside the image.
        # In which case we set the RA / DEC uncertainties to infinity.
        # The downside of this vectorized approach is that the position
        # errors for all the sources will be set to infinity.
        ra_errors = numpy.empty(num_islands).fill(numpy.inf)
        dec_errors = numpy.empty(num_islands).fill(numpy.inf)

    theta_celes_values = (numpy.degrees(moments_of_sources[:, 0, 6:7]) +
                          input_for_second_part[:, 2:3]) % 180

    theta_celes_errors = numpy.degrees(moments_of_sources[:, 1, 6:7])

    # This should also work for any nan value of theta_dc.
    theta_dc_celes_values = \
        (numpy.degrees(moments_of_sources[:, 0, 9:10]) +
         input_for_second_part[:, 2:3]) % 180

    theta_dc_celes_errors = numpy.degrees(moments_of_sources[:, 1, 9:10])

    try:
        end_smaj_ra_dec = wcs.all_p2s(input_for_second_part[:, [3, 5]])
        end_smin_ra_dec = wcs.all_p2s(input_for_second_part[:, [7, 9]])

    except RuntimeError:
        # Here we are again facing a downside of the vectorized
        # approach: if the FWHM ends of any source give a problem
        # wrt to pixel to sky conversion, all sky coordinates have to
        # be set to nans.
        logger.debug("pixel_to_spatial failed")
        end_smaj_ra_dec = \
            numpy.empty((num_islands, 2)).fill(numpy.nan)
        end_smin_ra_dec = \
            numpy.empty((num_islands, 2)).fill(numpy.nan)

    smaj_asec = numpy.empty(num_islands, dtype=numpy.float64)
    coordinates.angsep_vectorized(sky_barycenters, end_smaj_ra_dec,
                                  smaj_asec)

    scaling_smaj = smaj_asec / moments_of_sources[:, 0, 4]

    errsmaj_asec = scaling_smaj * moments_of_sources[:, 1, 4]

    smin_asec = numpy.empty(num_islands, dtype=numpy.float64)
    coordinates.angsep_vectorized(sky_barycenters, end_smin_ra_dec,
                                  smin_asec)

    scaling_smin = smin_asec / moments_of_sources[:, 0, 5]

    errsmin_asec = scaling_smin * moments_of_sources[:, 1, 5]

    return (moments_of_sources, sky_barycenters, xpositions,
            ypositions, ra_errors, dec_errors, error_radii, smaj_asec,
            errsmaj_asec, smin_asec, errsmin_asec, theta_celes_values,
            theta_celes_errors, theta_dc_celes_values, theta_dc_celes_errors)


@guvectorize([(int32, int32, int32[:], int32[:], int32, float32[:],
             float32[:, :])], '(), (), (n), (n), (), (k), (l, p)',
             nopython=True)
def calculate_Gaussian_islands(chunkposx, chunkposy, posx, posy, no_pixels,
                               gaussian_parms, islands_map):
    """Based on the derived Gaussian parameters, most likely through moments,
    reconstruct every island and add it to islands_map. Initially, this map
    will contain only zeroes. Note that the islands are reconstructed
    only where the island pixel value exceeded the local analysis threshold
    since this is how not only no_pixels, but also posx and posy - as positions
    relative to chunkposx and chunkposy - have been determined. The islands
    constructed in this manner can be used to derive a residual map.

    Args:
        chunkposx (numpy.ndarray): Row index of the top left corner of the
                                   rectangular slice encompassing the island
                                   relative to the top left corner of the image,
                                   which has pixel coordinates (0, 0), i.e. we
                                   need chunkposx to return to absolute pixel
                                   coordinates.

        chunkposy (numpy.ndarray): Column index of the top left corner of the
                                   rectangular slice encompassing the island
                                   relative to the top left corner of the image,
                                   which has pixel coordinates (0, 0), i.e. we
                                   need chunkposy to return to absolute pixel
                                   coordinates.

        posx (numpy.ndarray): Row indices of the pixels in island_data as taken
                              from the actual 2D images data (rectangular slice).

        posy (numpy.ndarray): Column indices of the pixels in island_data as
                              taken from the actual 2D images data (rectangular
                              slice).

        no_pixels (integer): The number of pixels that constitute the island.

        gaussian_parms (numpy.ndarray): an array of 6 floats containing
                                the derived Gaussian profile parameters of the
                                island, i.e. peak flux density, x barycenter,
                                y barycenter, semimajor axis, semiminor axis
                                and position angle. This derivation can be
                                either through moments calculation or through
                                fitting, in both cases one can calculate
                                residuals.

        islands_map (numpy.ndarray): Initially a 2D array with only zeroes with
                                     the same shape as the astronomical image
                                     that we are processing, i.e. the same shape
                                     as data_bgsubbed from the ImageData class
                                     instantiation. The Gaussian islands
                                     computed here are inserted (i.e. added)
                                     to this map.

    Returns:
        None (because of the guvectorize decorator), but islands_map - initially
        all zeroes - has reconstructed sources based on the Gaussian parameters
        derived from e.g. moments computations. These values are filled in where
        the source pixel values are above the local analysis threshold. At other
        pixel positions they remain zero.

    """
    peak = gaussian_parms[0]
    # xbar and ybar are relative to the upper left corner of the image, but for
    # this calculation we need the barycenter position relative to the upper
    # corner of the slice object encompassing the island, since also posx and
    # posy are defined in that manner.
    xbar = gaussian_parms[1] - chunkposx
    ybar = gaussian_parms[2] - chunkposy
    smaj = gaussian_parms[3]
    smin = gaussian_parms[4]
    theta = gaussian_parms[5]

    # Compute the residuals based on the derived Gaussian parameters.
    for index in range(no_pixels):
        # Calculate the position in " absolute" pixel coordinates, i.e.
        # relative to the upper left corner of the image, with row and column
        # index = 0.
        map_position = (posx[index] + chunkposx, posy[index] + chunkposy)
        islands_map[map_position] = peak * numpy.exp(-numpy.log(2) * (
                ((numpy.cos(theta) * (posx[index] - xbar)
                 + numpy.sin(theta) * (posy[index] - ybar)) / smin) ** 2 +
                ((numpy.cos(theta) * (posy[index] - ybar)
                 - numpy.sin(theta) * (posx[index] - xbar)) / smaj) ** 2))



"""
Generic utility routines for number handling and calculating (specific)
variances used by the TKP sourcefinder.
"""

import numpy
from numpy.ma import MaskedArray
from scipy.special import erf
from scipy.optimize import fsolve


# CODE & NUMBER HANDLING ROUTINES
#
def find_true_std(sigma, clip_limit, clipped_std):
    help1 = clip_limit / (sigma * numpy.sqrt(2))
    help2 = numpy.sqrt(2 * numpy.pi) * erf(help1)
    return (sigma ** 2 * (help2 - 2 * numpy.sqrt(2) * help1 *
            numpy.exp(-help1 ** 2)) - clipped_std ** 2 * help2)


def indep_pixels(n, correlation_lengths):
    corlengthlong, corlengthshort = correlation_lengths
    correlated_area = 0.25 * numpy.pi * corlengthlong * corlengthshort
    return n / correlated_area


def sigma_clip(data, kappa=2.0, max_iter=100,
               centref=numpy.median, distf=numpy.var, my_iterations=0, limit=None):
    """Iterative clipping

    By default, this performs clipping of the standard deviation about the
    median of the data. But by tweaking centref/distf, it could be much
    more general.

    max_iter sets the maximum number of iterations used.

    my_iterations is a counter for recursive operation of the code; leave it
    alone unless you really want to pretend to jump into the middle of a loop.

    sigma is subtle: if a callable is given, it is passed a copy of the data
    array and can calculate a clipping limit. See, for e.g., unbiased_sigma()
    defined above. However, if it isn't callable, sigma is assumed to just set
    a hard limit.

    To do: Improve documentation
            -Returns???
            -How does it make use of the beam? (It estimates the noise correlation)
    """
    if my_iterations >= max_iter:
        # Exceeded maximum number of iterations; return
        return data, my_iterations

    # Numpy 1.1 breaks std() for MaskedArray: see
    # <http://www.scipy.org/scipy/numpy/wiki/MaskedArray>.
    # MaskedArray.compressed() returns a 1-D array of non-masked data.
    if isinstance(data, MaskedArray):
        data = data.compressed()
    centre = centref(data)
    n = numpy.size(data)
    if n < 1:
        # This chunk is too small for processing; return an empty array.
        return numpy.array([]), 0, 0, 0

    clipped_var = distf(data)

    std = numpy.sqrt(clipped_var)

    if limit is not None:
        std_corr_for_clipping_bias = fsolve(find_true_std, std,
                                            args=(limit, std))[0]
    else:
        std_corr_for_clipping_bias = std

    limit = kappa * std_corr_for_clipping_bias

    newdata = data.compress(abs(data - centre) <= limit)

    if len(newdata) != len(data) and len(newdata) > 0:
        my_iterations += 1
        return sigma_clip(newdata, kappa, max_iter, centref, distf,
                          my_iterations, limit=limit)
    else:
        return newdata, std_corr_for_clipping_bias, centre, my_iterations



import logging

import numpy
from math import degrees, sqrt, sin, pi, cos
from dataclasses import dataclass
from sourcefinder.utility.coordinates import WCS

logger = logging.getLogger(__name__)


@dataclass
class DataAccessor:
    # Sphinx only picks up the class docstring if it's under an __init__
    # *le sigh*
    """
    Base class for accessors used with
    :class:`sourcefinder.image.ImageData`.

    Data accessors provide a uniform way for the ImageData class (ie,
    generic image representation) to access the various ways in which
    images may be stored (FITS files, arrays in memory, potentially HDF5,
    etc).

    This class cannot be instantiated directly, but should be subclassed
    and the abstract properties provided. Note that all abstract
    properties are required to provide a valid accessor.

    Additional properties may also be provided by subclasses. However,
    TraP components are required to degrade gracefully in the absence of
    this optional properties.

    The required attributes are as follows:

    Attributes:
        beam (tuple): Restoring beam. Tuple of three floats:
            semi-major axis (in pixels), semi-minor axis (pixels)
            and position angle (radians).
        centre_ra (float): Right ascension at the central pixel of the image.
            Units of J2000 decimal degrees.
        centre_decl (float): Declination at the central pixel of the image.
            Units of J2000 decimal degrees.
        data(numpy.ndarray): Two dimensional numpy.ndarray of floating point
            pixel values.
            (TODO: Definitive statement on orientation/transposing.)
        freq_bw(float): The frequency bandwidth of this image in Hz.
        freq_eff(float): Effective frequency of the image in Hz.
            That is, the mean frequency of all the visibility data which
            comprises this image.
        pixelsize(tuple): (x, y) tuple representing the size of a pixel
            along each axis in units of degrees.
        tau_time(float): Total time on sky in seconds.
        taustart_ts(float): Timestamp of the first integration which
            constitutes part of this image. MJD in seconds.
        url(string): A (string) URL representing the location of the image
            at time of processing.
        wcs(:class:`sourcefinder.utility.coordinates.WCS`): An instance of
            :py:class:`sourcefinder.utility.coordinates.WCS`,
            describing the mapping from data pixels to sky-coordinates.

    The class also provides some common functionality:
    static methods used for parsing datafiles, and an 'extract_metadata'
    function which provides key info in a simple dict format.
    """

    beam: tuple
    centre_ra: float
    centre_decl: float
    data: numpy.ndarray
    freq_bw: float
    freq_eff: float
    pixelsize: tuple
    tau_time: float
    taustart_ts: float
    url: str
    wcs: WCS

    def extract_metadata(self) -> dict:
        """
        Massage the class attributes into a flat dictionary with
        database-friendly values.

        While rather tedious, this is easy to serialize and store separately
        to the actual image data.

        May be extended by subclasses to return additional data.
        """
        return {
            'tau_time': self.tau_time,
            'freq_eff': self.freq_eff,
            'freq_bw': self.freq_bw,
            'taustart_ts': self.taustart_ts,
            'url': self.url,
            'beam_smaj_pix': self.beam[0],
            'beam_smin_pix': self.beam[1],
            'beam_pa_rad': self.beam[2],
            'centre_ra': self.centre_ra,
            'centre_decl': self.centre_decl,
            'deltax': self.pixelsize[0],
            'deltay': self.pixelsize[1],
        }

    def parse_pixelsize(self):
        """

        Returns:
          - deltax: pixel size along the x axis in degrees
          - deltay: pixel size along the x axis in degrees

        """
        wcs = self.wcs
        # Check that pixels are square
        # (Would have to be pretty strange data for this not to be the case)
        assert wcs.cunit[0] == wcs.cunit[1]
        if wcs.cunit[0] == "deg":
            deltax = wcs.cdelt[0]
            deltay = wcs.cdelt[1]
        elif wcs.cunit[0] == "rad":
            deltax = degrees(wcs.cdelt[0])
            deltay = degrees(wcs.cdelt[1])
        else:
            raise ValueError("Unrecognised WCS co-ordinate system")

        # NB. What's a reasonable epsilon here?
        eps = 1e-7
        if abs(abs(deltax) - abs(deltay)) > eps:
            raise ValueError("Image WCS header suggests non-square pixels."
                             "This is an untested use case, and may break things -"
                             "specifically the skyregion tracking but possibly other stuff too.")
        return deltax, deltay

    @staticmethod
    def degrees2pixels(bmaj, bmin, bpa, deltax, deltay):
        """
        Convert beam in degrees to beam in pixels and radians.
        For example Fits beam parameters are in degrees.

        Arguments:
          - bmaj:   Beam major axis in degrees
          - bmin:   Beam minor axis in degrees
          - bpa:    Beam position angle in degrees
          - deltax: Pixel size along the x axis in degrees
          - deltay: Pixel size along the y axis in degrees

        Returns:
          - semimaj: Beam semi-major axis in pixels
          - semimin: Beam semi-minor axis in pixels
          - theta:   Beam position angle in radians
        """
        theta = pi * bpa / 180
        semimaj = (bmaj / 2.) * (sqrt(
            (sin(theta) ** 2) / (deltax ** 2) +
            (cos(theta) ** 2) / (deltay ** 2))
        )
        semimin = (bmin / 2.) * (sqrt(
            (cos(theta) ** 2) / (deltax ** 2) +
            (sin(theta) ** 2) / (deltay ** 2))
        )
        return (semimaj, semimin, theta)



"""
Data accessors.

These can be used to populate ImageData objects based on some data source
(FITS file, array in memory... etc).
"""

import os

import astropy.io.fits as pyfits
from astropy.io.fits.hdu.hdulist import HDUList

from sourcefinder.accessors import detection
from sourcefinder.accessors.fitsimageblob import FitsImageBlob
from sourcefinder.accessors.lofarfitsimage import LofarFitsImage
from sourcefinder.image import ImageData


def sourcefinder_image_from_accessor(image, **args):
    """Create a source finder ImageData object from an image 'accessor'

    Args:

        - image (DataAccessor): FITS/AIPS/HDF5 image available through
          an accessor.

    Returns:
        (:class:`sourcefinder.image.ImageData`): a source finder image.
    """
    image = ImageData(image.data, image.beam, image.wcs, **args)
    return image


def writefits(data, filename, header={}):
    """
    Dump a NumPy array to a FITS file.

    Key/value pairs for the FITS header can be supplied in the optional
    header argument as a dictionary.
    """
    if header.__class__.__name__ == 'Header':
        pyfits.writeto(filename, data.transpose(), header)
    else:
        hdu = pyfits.PrimaryHDU(data.transpose())
        for key in header.keys():
            hdu.header.update(key, header[key])
        hdu.writeto(filename)


def open(path, *args, **kwargs):
    """
    Returns an accessor object (if available) for the file or directory 'path'.

    We try all the possible accessors in order from most specific to least
    specific. That is, if possible, we prefer an accessor providing
    LofarAccessor to one providing DataAccessor, but we accept the latter if
    that's the only possible match.

    Will raise an exception if something went wrong or no matching accessor
    class is found.
    """
    if type(path) == HDUList:
        return FitsImageBlob(path, *args, **kwargs)
    elif type(path) == str:
        if not os.access(path, os.F_OK):
            raise OSError("%s does not exist!" % path)
        if not os.access(path, os.R_OK):
            raise OSError("Don't have permission to read %s!" % path)
        Accessor = detection.detect(path)
        if not Accessor:
            raise OSError("no accessor found for %s" % path)
        return Accessor(path, *args, **kwargs)
    else:
        raise Exception("image should be path or HDUlist, got " + str(path))



import json
import logging


def parse_monitoringlist_positions(args, str_name="monitor_coords",
                                   list_name="monitor_list"):
    """Loads a list of monitoringlist (RA,Dec) tuples from cmd line args object.

    Processes the flags "--monitor-coords" and "--monitor-list"
    NB This is just a dumb function that does not care about units,
    those should be matched against whatever uses the resulting values...
    """
    monitor_coords = []
    if hasattr(args, str_name) and getattr(args, str_name):
        try:
            monitor_coords.extend(json.loads(getattr(args, str_name)))
        except ValueError:
            logging.error("Could not parse monitor-coords from command line:"
                          "string passed was:\n%s" % (getattr(args, str_name),)
                          )
            raise
    if hasattr(args, list_name) and getattr(args, list_name):
        try:
            mon_list = json.load(open(getattr(args, list_name)))
            monitor_coords.extend(mon_list)
        except ValueError:
            logging.error("Could not parse monitor-coords from file: "
                          + getattr(args, list_name))
            raise
    return monitor_coords



import numpy
import pytz
import datetime
import dateutil.parser
import logging
import re

import astropy.io.fits as pyfits

from sourcefinder.accessors.dataaccessor import DataAccessor
from sourcefinder.utility.coordinates import WCS

logger = logging.getLogger(__name__)


class FitsImage(DataAccessor):
    """
    Use PyFITS to pull image data out of a FITS file.

    Provide standard attributes, as per :class:`DataAccessor`. In addition, we
    provide a ``telescope`` attribute if the FITS file has a ``TELESCOP``
    header.
    """
    def __init__(self, url, plane=None, beam=None, hdu_index=0):
        self.url = url
        self.header = self._get_header(hdu_index)
        self.wcs = self.parse_coordinates()
        self.data = self.read_data(hdu_index, plane)
        self.taustart_ts, self.tau_time = self.parse_times()
        self.freq_eff, self.freq_bw = self.parse_frequency()
        self.pixelsize = self.parse_pixelsize()
        if beam:
            (bmaj, bmin, bpa) = beam
        else:
            (bmaj, bmin, bpa) = self.parse_beam()
        self.beam = self.degrees2pixels(
            bmaj, bmin, bpa, self.pixelsize[0], self.pixelsize[1]
        )
        self.centre_ra, self.centre_decl = self.calculate_phase_centre()

        # Bonus attribute
        if 'TELESCOP' in self.header:
            self.telescope = self.header['TELESCOP']

    def _get_header(self, hdu_index):
        with pyfits.open(self.url) as hdulist:
            hdu = hdulist[hdu_index]
        return hdu.header.copy()

    def read_data(self, hdu_index, plane):
        """
        Read and store data from our FITS file.

        NOTE: PyFITS reads the data into an array indexed as [y][x]. We
        take the transpose to make this more intuitively reasonable and
        consistent with (eg) ds9 display of the FitsFile. Transpose back
        before viewing the array with RO.DS9, saving to a FITS file,
        etc.
        """
        with pyfits.open(self.url) as hdulist:
            hdu = hdulist[hdu_index]
            data = numpy.float64(hdu.data.squeeze())
        if plane is not None and len(data.shape) > 2:
            data = data[plane].squeeze()
        n_dim = len(data.shape)
        if n_dim != 2:
            logger.warning((f"Loaded datacube with {n_dim:d} dimensions, "
                             "assuming Stokes I and taking plane 0."))
            data = data[0, :, :]
        data = data.transpose()
        return data

    def parse_coordinates(self):
        """Returns a WCS object"""
        header = self.header
        wcs = WCS()
        try:
            wcs.crval = header['crval1'], header['crval2']
            wcs.crpix = header['crpix1'] - 1, header['crpix2'] - 1
            wcs.cdelt = header['cdelt1'], header['cdelt2']
        except KeyError:
            msg = "Coordinate system not specified in FITS"
            logger.error(msg)
            raise TypeError(msg)
        try:
            wcs.ctype = header['ctype1'], header['ctype2']
        except KeyError:
            wcs.ctype = 'unknown', 'unknown'
        try:
            wcs.crota = float(header['crota1']), float(header['crota2'])
        except KeyError:
            wcs.crota = 0., 0.
        try:
            wcs.cunit = header['cunit1'], header['cunit2']
        except KeyError:
            # The "Definition of the Flexible Image Transport System", version
            # 3.0, tells us that "units for celestial coordinate systems defined
            # in this Standard must be degrees", so we assume that if nothing else
            # is specifiedj
            msg = "WCS units unknown; using degrees"
            logger.warning(msg)
            wcs.cunit = 'deg', 'deg'
        return wcs


    def calculate_phase_centre(self):
        x, y = self.data.shape
        centre_ra, centre_decl = self.wcs.p2s((x / 2, y / 2))
        return float(centre_ra), float(centre_decl)


    def parse_frequency(self):
        """
        Set some 'shortcut' variables for access to the frequency parameters
        in the FITS file header.

        @param hdulist: hdulist to parse
        @type hdulist: hdulist
        """
        freq_eff = None
        freq_bw = None
        try:
            header = self.header
            if 'RESTFRQ' in header:
                freq_eff = header['RESTFRQ']
                if 'RESTBW' in header:
                    freq_bw = header['RESTBW']
                else:
                    logger.warning("bandwidth header missing in image {},"
                                   " setting to 1 MHz".format(self.url))
                    freq_bw = 1e6
            elif ('CTYPE3' in header) and (header['CTYPE3'] in ('FREQ', 'VOPT')):
                freq_eff = header['CRVAL3']
                freq_bw = header['CDELT3']
            elif ('CTYPE4' in header) and (header['CTYPE4'] in ('FREQ', 'VOPT')):
                freq_eff = header['CRVAL4']
                freq_bw = header['CDELT4']
            else:
                freq_eff = header['RESTFREQ']
                freq_bw = 1e6
        except KeyError:
            msg = "Frequency not specified in headers for {}".format(self.url)
            logger.error(msg)
            raise TypeError(msg)

        return freq_eff, freq_bw


    def parse_beam(self):
        """Read and return the beam properties bmaj, bmin and bpa values from
        the fits header.

        Returns:
          - Beam parameters, (semimajor, semiminor, position angle)
            in (pixels, pixels, radians)
        """
        # AIPS FITS file; stored in the history section
        beam_regex = re.compile(r'''
            BMAJ
            \s*=\s*
            (?P<bmaj>[-\d\.eE]+)
            \s*
            BMIN
            \s*=\s*
            (?P<bmin>[-\d\.eE]+)
            \s*
            BPA
            \s*=\s*
            (?P<bpa>[-\d\.eE]+)
            ''', re.VERBOSE)

        bmaj, bmin, bpa = None, None, None
        header = self.header
        try:
            # MIRIAD FITS file
            bmaj = header['BMAJ']
            bmin = header['BMIN']
            bpa = header['BPA']
        except KeyError:

            def get_history(hdr):
                """
                Returns all history cards in FITS header hdr as a list of strings.
                """
                return hdr['HISTORY']

            for hist_entry in get_history(header):
                results = beam_regex.search(hist_entry)
                if results:
                    bmaj, bmin, bpa = [float(results.group(key)) for
                                       key in ('bmaj', 'bmin', 'bpa')]
                    break

        return bmaj, bmin, bpa


    def parse_times(self):
        """Returns:
          - taustart_ts: tz naive (implicit UTC) datetime at start of observation.
          - tau_time: Integration time, in seconds
        """
        # Attempt to do something sane with timestamps.
        try:
            start = self.parse_start_time()
        except KeyError:
            # If no start time specified, give up:
            logger.warning("Timestamp not specified in FITS file:"
                        " using 'now' with dummy (zero-valued) integration time.")
            return datetime.datetime.now(), 0.

        try:
            end = dateutil.parser.parse(self.header['end_utc'])
        except KeyError:
            msg = "End time not specified in {}, setting to start".format(
                self.url)
            logger.warning(msg)
            end = start

        delta = end - start
        tau_time = delta.total_seconds()

        # For simplicity, the database requires naive datetimes (implicit UTC)
        # So we convert to UTC and then drop the timezone:
        try:
            timezone = pytz.timezone(self.header['timesys'])
            start_w_tz = start.replace(tzinfo=timezone)
            start_utc = pytz.utc.normalize(start_w_tz.astimezone(pytz.utc))
            return start_utc.replace(tzinfo=None), tau_time
        except (pytz.UnknownTimeZoneError, KeyError):
            logger.debug("Timezone not specified in FITS file: assuming UTC.")
            return start, tau_time


    def parse_start_time(self):
        """
        Returns:
          - start time of image as an instance of ``datetime.datetime``
        """
        header = self.header
        try:
            if ";" in header['date-obs']:
                start = dateutil.parser.parse(header['date-obs']\
                    .split(";")[0].split('"')[1])
            else:
                start = dateutil.parser.parse(header['date-obs'])
        except AttributeError:
            # Maybe it's a float, Westerbork-style?
            if isinstance(header['date-obs'], float):
                logger.warning("Non-standard date specified in FITS file!")
                frac, year = numpy.modf(header['date-obs'])
                start = datetime.datetime(int(year), 1, 1)
                delta = datetime.timedelta(365.242199 * frac)
                start += delta
            else:
                raise KeyError("Timestamp in fits file unreadable")
        return start



from dataclasses import dataclass


@dataclass
class LofarAccessor():
    """
    Additional metadata required for processing LOFAR images through QC
    checks.

    Attributes:
        antenna_set (string): Antenna set in use during observation.
            String; 'LBA_INNER', 'LBA_OUTER', 'LBA_SPARSE', 'LBA' or 'HBA'
        ncore(int): Number of core stations in use during observation.
        nremote(int): Number of remote stations in use during observation.
        nintl(int): Number of international stations in use during observation.
        subbandwidth(float): Width of a subband in Hz.
        subbands(int): Number of subbands.
    """

    antenna_set: str
    ncore: int
    nremote: int
    nintl: int
    subbandwidth: float
    subbands: int



import logging

import numpy

from sourcefinder.accessors.fitsimage import FitsImage

logger = logging.getLogger(__name__)


class FitsImageBlob(FitsImage):
    """
    A Fits image Blob. Same as ``sourcefinder.accessors.fitsimage.FitsImage`` but
    constructed from an in memory fits file, not a fits file on disk.
    """

    def __init__(self, hdulist, plane=None, beam=None, hdu_index=0):
        # set the URL in case we need it during header parsing for error loggign
        self.url = "AARTFAAC streaming image"
        # super(FitsImage, self).__init__()

        self.header = self._get_header(hdulist, hdu_index)
        self.wcs = self.parse_coordinates()
        self.data = self.read_data(hdulist, hdu_index, plane)
        self.taustart_ts, self.tau_time = self.parse_times()
        self.freq_eff, self.freq_bw = self.parse_frequency()
        self.pixelsize = self.parse_pixelsize()

        elements = "memory://AARTFAAC", self.taustart_ts, self.tau_time, \
                   self.freq_eff, self.freq_bw
        self.url = "_".join([str(x) for x in elements])

        if beam:
            (bmaj, bmin, bpa) = beam
        else:
            (bmaj, bmin, bpa) = self.parse_beam()
        self.beam = self.degrees2pixels(
            bmaj, bmin, bpa, self.pixelsize[0], self.pixelsize[1]
        )
        self.centre_ra, self.centre_decl = self.calculate_phase_centre()

        # Bonus attribute
        if 'TELESCOP' in self.header:
            self.telescope = self.header['TELESCOP']

    def _get_header(self, *args):
        hdulist = args[0]
        hdu_index = args[1]
        return hdulist[hdu_index].header

    def read_data(self, *args):
        hdulist = args[0]
        hdu_index = args[1]
        plane = args[2]
        hdu = hdulist[hdu_index]
        data = numpy.float64(hdu.data.squeeze())
        if plane is not None and len(data.shape) > 2:
            data = data[plane].squeeze()
        n_dim = len(data.shape)
        if n_dim != 2:
            logger.warning(
                "Loaded datacube with %s dimensions, assuming Stokes I and taking plane 0" % n_dim)
            data = data[0, :, :]
        data = data.transpose()
        return data



"""
This module implements the CASA LOFAR data container format, described in this
document:

http://www.lofar.org/operations/lib/exe/fetch.php?media=:public:documents:casa_image_for_lofar_0.03.00.pdf
"""
import datetime
import logging

import numpy
from casacore.tables import table as casacore_table

from sourcefinder.accessors.casaimage import CasaImage
from sourcefinder.accessors.lofaraccessor import LofarAccessor
from sourcefinder.utility.coordinates import julian2unix

logger = logging.getLogger(__name__)

subtable_names = (
    'LOFAR_FIELD',
    'LOFAR_ANTENNA',
    'LOFAR_HISTORY',
    'LOFAR_ORIGIN',
    'LOFAR_QUALITY',
    'LOFAR_STATION',
    'LOFAR_POINTING',
    'LOFAR_OBSERVATION'
)


class LofarCasaImage(CasaImage, LofarAccessor):
    """
    Use casacore to pull image data out of an Casa table.

    This accessor assumes the casatable contains the values described in the
    CASA Image description for LOFAR. 0.03.00.

    Args:
      - url: location of CASA table
      - plane: if datacube, what plane to use
      - beam: (optional) beam parameters in degrees, in the form
        (bmaj, bmin, bpa). Will attempt to read from header if
        not supplied.
    """

    def __init__(self, url, plane=0, beam=None):
        super().__init__(url, plane, beam)
        table = casacore_table(self.url, ack=False)
        subtables = self.open_subtables(table)
        self.taustart_ts = self.parse_taustartts(subtables)
        self.tau_time = self.parse_tautime(subtables)

        # Additional, LOFAR-specific metadata
        self.antenna_set = self.parse_antennaset(subtables)
        self.ncore, self.nremote, self.nintl = self.parse_stations(subtables)
        self.subbandwidth = self.parse_subbandwidth(subtables)
        self.subbands = self.parse_subbands(subtables)

    def open_subtables(self, table):
        """open all subtables defined in the LOFAR format
        args:
            table: a casacore table handler to a LOFAR CASA table
        returns:
            a dict containing all LOFAR CASA subtables
        """
        subtables = {}
        for subtable in subtable_names:
            subtable_location = table.getkeyword("ATTRGROUPS")[subtable]
            subtables[subtable] = casacore_table(subtable_location, ack=False)
        return subtables

    def parse_taustartts(self, subtables):
        """ extract image start time from CASA table header
        """
        # Note that we sort the table in order of ascending start time then
        # choose the first value to ensure we get the earliest possible
        # starting time.
        observation_table = subtables['LOFAR_OBSERVATION']
        julianstart = observation_table.query(
            sortlist="OBSERVATION_START", limit=1).getcell(
            "OBSERVATION_START", 0
        )
        unixstart = julian2unix(julianstart)
        taustart_ts = datetime.datetime.fromtimestamp(unixstart)
        return taustart_ts

    @staticmethod
    def non_overlapping_time(series):
        """
        Returns the sum of total ranges without overlap.

        series: a list of 2 item tuples representing ranges.
        """
        series.sort()
        overlap = total = 0
        for n, (start, end) in enumerate(series):
            total += end - start
            for (nextstart, nextend) in series[n + 1:]:
                if nextstart >= end:
                    break
                overlapstart = max(nextstart, start)
                overlapend = min(nextend, end)
                overlap += overlapend - overlapstart
                start = overlapend
        return total - overlap

    def parse_tautime(self, subtables):
        """
        Returns the total on-sky time for this image.
        """
        origin_table = subtables['LOFAR_ORIGIN']
        startcol = origin_table.col('START')
        endcol = origin_table.col('END')
        series = [(int(start), int(end)) for start, end in
                  zip(startcol, endcol)]
        tau_time = LofarCasaImage.non_overlapping_time(series)
        return tau_time

    def parse_antennaset(self, subtables):
        observation_table = subtables['LOFAR_OBSERVATION']
        antennasets = CasaImage.unique_column_values(observation_table,
                                                     "ANTENNA_SET")
        if len(antennasets) == 1:
            return antennasets[0]
        else:
            raise Exception("Cannot handle multiple antenna sets in image")

    def parse_subbands(self, subtables):
        origin_table = subtables['LOFAR_ORIGIN']
        num_chans = CasaImage.unique_column_values(origin_table, "NUM_CHAN")
        if len(num_chans) == 1:
            return num_chans[0]
        else:
            raise Exception(
                "Cannot handle varying numbers of channels in image")

    def parse_subbandwidth(self, subtables):
        # subband
        # see http://www.lofar.org/operations/doku.php?id=operator:background_to_observations&s[]=subband&s[]=width&s[]=clock&s[]=frequency
        freq_units = {
            'Hz': 1,
            'kHz': 10 ** 3,
            'MHz': 10 ** 6,
            'GHz': 10 ** 9,
        }
        observation_table = subtables['LOFAR_OBSERVATION']
        clockcol = observation_table.col('CLOCK_FREQUENCY')
        clock_values = CasaImage.unique_column_values(observation_table,
                                                      "CLOCK_FREQUENCY")
        if len(clock_values) == 1:
            clock = clock_values[0]
            unit = clockcol.getkeyword('QuantumUnits')[0]
            trueclock = freq_units[unit] * clock
            subbandwidth = trueclock / 1024
            return subbandwidth
        else:
            raise Exception("Cannot handle varying clocks in image")

    def parse_stations(self, subtables):
        """Extract number of specific LOFAR stations used
        returns:
            (number of core stations, remote stations, international stations)
        """

        observation_table = subtables['LOFAR_OBSERVATION']
        antenna_table = subtables['LOFAR_ANTENNA']
        nvis_used = observation_table.getcol('NVIS_USED')
        names = numpy.array(antenna_table.getcol('NAME'))
        mask = numpy.sum(nvis_used, axis=2) > 0
        used = names[mask[0]]
        ncore = nremote = nintl = 0
        for station in used:
            if station.startswith('CS'):
                ncore += 1
            elif station.startswith('RS'):
                nremote += 1
            else:
                nintl += 1
        return ncore, nremote, nintl



"""
Gaussian deconvolution.
"""

from math import sin, cos, atan, sqrt, pi
from numba import njit

@njit
def deconv(fmaj, fmin, fpa, cmaj, cmin, cpa):
    """
    Deconvolve a Gaussian "beam" from a Gaussian component.

    When we fit an elliptical Gaussian to a point in our image, we are
    actually fitting to a convolution of the physical shape of the source with
    the beam pattern of our instrument. This results in the fmaj/fmin/fpa
    arguments to this function.

    Since the shape of the (clean) beam (arguments cmaj/cmin/cpa) is known, we
    can deconvolve it from the fitted parameters to get the "real" underlying
    physical source shape, which is what this function returns.

    Args:
        fmaj (float): Fitted major axis
        fmin (float): Fitted minor axis
        fpa (float):  Fitted position angle of major axis
        cmaj (float): Clean beam major axis
        cmin (float): Clean beam minor axis
        cpa (float):  Clean beam position angle of major axis

    Returns:
        tuple: real major axis, real minor axis, real position angle of major
               axis, number of components which failed to deconvolve
    """
    HALF_RAD = 90.0 / pi
    cmaj2 = cmaj * cmaj
    cmin2 = cmin * cmin
    fmaj2 = fmaj * fmaj
    fmin2 = fmin * fmin
    theta = (fpa - cpa) / HALF_RAD
    det = ((fmaj2 + fmin2) - (cmaj2 + cmin2)) / 2.0
    rhoc = (fmaj2 - fmin2) * cos(theta) - (cmaj2 - cmin2)
    sigic2 = 0.0
    rhoa = 0.0
    ierr = 0

    if abs(rhoc) > 0.0:
        sigic2 = atan((fmaj2 - fmin2) * sin(theta) / rhoc)
        rhoa = (((cmaj2 - cmin2) - (fmaj2 - fmin2) * cos(theta)) /
                (2.0 * cos(sigic2)))

    rpa = sigic2 * HALF_RAD + cpa
    rmaj = det - rhoa
    rmin = det + rhoa

    if rmaj < 0:
        ierr += 1
        rmaj = 0
    if rmin < 0:
        ierr += 1
        rmin = 0

    rmaj = sqrt(rmaj)
    rmin = sqrt(rmin)
    if rmaj < rmin:
        rmaj, rmin = rmin, rmaj
        rpa += 90

    rpa = (rpa + 900) % 180
    if not abs(rmaj):
        rpa = 0.0
    elif not abs(rmin) and (45.0 < abs(rpa - fpa) < 135.0):
        rpa = (rpa + 450.0) % 180.0

    return rmaj, rmin, rpa, ierr
